/***********************************************************************

	File:	bb.c
	Rev:	b-1
	Date:	11/14/96

	Copyright (c) 1995, 1999 by David M. Warme

************************************************************************

	The main routine for the branch-and-cut.  (Actually the name
	"bc" is taken so I use "bb" for branch-and-bound.)  It takes
	the phase 1 output from the "prep" program and finds the
	Steiner minimal tree as a minimal spanning subset of full sets.

************************************************************************

	Modification Log:

	a-1:	11/17/95	warme
		: Created.
	b-1:	11/14/96	warme
		: Renamed this program.
		: Split off the cutset stuff into another file.
		: Other cleanups for release.

************************************************************************/

#include "bb.h"
#include "config.h"
#include "constrnt.h"
#include "cutset.h"
#include "genps.h"
#include "p1io.h"
#include "sec2.h"
#include "sec_comp.h"
#include "sec_heur.h"
#include "steiner.h"
#include "ub.h"

#include <signal.h>


/*
 * Global Routines
 */

int			main (int, char **);

void			new_upper_bound (double, struct bbinfo *);


bool			Choose_Branch_Vars_Carefully = TRUE;
bool			Print_Root_LP = FALSE;

bool			Seed_Pool_With_2SECs = TRUE;

volatile bool		force_branch_flag;

#if defined(CPLEX) AND (CPLEX >= 40)
CPXENVptr		cplex_env;
#endif

#ifdef LPSOLVE
bool			Use_Perturbations = FALSE;
bool			Use_Scaling = FALSE;
#endif

/*
 * External References
 */

	/* None */


/*
 * Local Equates
 */


	/* Lower bound outcomes... */
#define	LB_INFEASIBLE		1
#define	LB_CUTOFF		2
#define	LB_INTEGRAL		3
#define	LB_FRACTIONAL		4
#define LB_PREEMPTED		5

	/* Variable fixing outcomes... */
#define	VFIX_NOTHING_FIXED	0
#define	VFIX_VARIABLES_FIXED	1
#define	VFIX_INFEASIBLE		2

#define	UP_FIRST	TRUE

	/* Next node policies... */
#define	NN_DEPTH_FIRST		0
#define	NN_BEST_NODE		1


/*
 * Local Types
 */

#ifdef CPLEX

struct basis_save {	/* Structure to save basis state for CPLEX */
	int *		cstat;
	int *		rstat;
};

static void		destroy_LP_basis (struct basis_save *);
static double		try_branch (LP_t *,
				    int,
				    int,
				    double *,
				    double,
				    struct basis_save *);
static void		save_LP_basis (LP_t *, struct basis_save *);
#endif


/*
 * Local Routines
 */

static void		add_bbnode (struct bbinfo *, int, int, double);
static void		bbheap_delete (struct bbnode *, struct bbtree *, int);
static void		bbheap_free (struct bbheap *);
static void		bbheap_init (struct bbheap *, bbheap_func_t *);
static void		bbheap_insert (struct bbnode *, struct bbtree *, int);
static dist_t		branch_and_cut (struct cinfo *,
					struct bbstats *,
					bitmap_t *);
static int		carefully_choose_branching_variable (struct bbinfo *,
							     double *,
							     double *);
static void		change_var_bounds (LP_t *, int, double, double);
static bool		check_for_better_IFS (double *,
					      struct bbinfo *,
					      double *);
static void		check_root_constraints (struct bbinfo *);
static int		choose_branching_variable (struct bbinfo *,
						   double *,
						   double *);
static int		compute_good_lower_bound (struct bbinfo *);
static void		cut_off_existing_nodes (double, struct bbtree *);
static void		decode_params (int, char **);
static void		delete_node_from_bbtree (struct bbnode *,
						 struct bbtree *);
static struct constraint * do_separations (struct bbinfo *,
					   cpu_time_t **);
static int		find_comps (bitmap_t *, bitmap_t *, struct cinfo *);
static int		fix_variables (struct bbinfo *,
				       int *, int,
				       int *, int);
static RETSIGTYPE	force_branch_signal_handler (int);
static bool		integer_feasible_solution (double *,
						   bitmap_t *,
						   bitmap_t *,
						   struct cinfo *,
						   int *);
static void		new_lower_bound (double, struct bbinfo *);
static int		node_is_better (struct bbnode *, struct bbnode *);
static int		node_is_worse (struct bbnode *, struct bbnode *);
static void		print_root_lp (struct bbinfo *);
static int		reduced_cost_var_fixing (struct bbinfo *);
static struct bbnode *	select_next_node (struct bbtree *);
static void		test_branching_variables (struct bbinfo *);
static void		trace_node (struct bbinfo *, char, char *);
static void		update_lp_solution_history (struct bbinfo *);
static void		update_node_preempt_value (struct bbinfo *);
static void		usage (void);


/*
 * Local Variables
 */

static bool		Check_Root_Constraints = FALSE;
static double		Initial_Upper_Bound = DBL_MAX;
static double		Trial_Branch_Threshold = DBL_MAX;

static int32u		cpu_time_limit;
static char *		me;
static FILE *		rcfile;

/*
 * The main routine for the "bb" program.  It takes the output from
 * the "prep" program (phase 1 of our Steiner tree program) and uses
 * a branch-and-cut to find the Steiner minimal tree.
 */

	int
main (

int		argc,
char **		argv
)
{
int			i;
int			j;
int			nmasks;
bitmap_t *		smt_mask;
struct cinfo		cinfo;
dist_t			length;
char *			descr;
struct pset *		pts;
char *			cp1;
char *			cp2;
char			c;
int *			edge_count;
int			smt_edge_count;
int			total_edge_count;
int			edge10;
int			max_edge_size;
int			nt;
double			gap;
double			redmst;
char			buf1 [32];
char			buf2 [32];
char			buf3 [32];
char			buf4 [32];
char			title [80];
struct bbstats		stats;

	setbuf (stdout, NULL);

	decode_params (argc, argv);

	init_tables ();

	read_phase_1_data (&cinfo);

	stats.p1time = cinfo.p1time;
	convert_cpu_time (stats.p1time, buf1);
	printf (" %% Phase 1: %s seconds\n", buf1);

	define_Plot_Terminals (cinfo.pts, cinfo.scale_value);

	nmasks = cinfo.num_edge_masks;

	smt_mask = NEWA (nmasks, bitmap_t);
	for (i = 0; i < nmasks; i++) {
		smt_mask [i] = 0;
	}

	/* Enable CPU time limitation, if any. */
	start_limiting_cpu_time (cpu_time_limit);

#ifdef CPLEX
 #if CPLEX >= 40
	/* Newer versions of CPLEX need to be initialized... */
	{ int status;
	  char msg [512];
	  FILE *	fp;
	  CPXCHANNELptr	_res, _warn, _error, _log;

	  cplex_env = CPXopenCPLEX (&status);
	  if (cplex_env EQ NULL) {
		if (CPXgeterrorstring (NULL, status, msg) EQ NULL) {
			strcpy (msg, "No CPLEX error message.");
		}
		fprintf (stderr, "%s\n", msg);
		goto shutdown;
	  }

	  /* Get rid of CPLEX's default message destinations. */
	  CPXgetchannels (cplex_env, &_res, &_warn, &_error, &_log);
	  CPXdisconnectchannel (cplex_env, _res);
	  CPXdisconnectchannel (cplex_env, _warn);
	  CPXdisconnectchannel (cplex_env, _error);
	  CPXdisconnectchannel (cplex_env, _log);

	  CPXsetintparam (cplex_env, CPX_PARAM_SCRIND, 0);

	  fp = fopen ("cplex.log", "a");
	  if (fp EQ NULL) {
		perror ("cplex.log");
shutdown:
		/* Shut down CPLEX and get out. */
		CPXcloseCPLEX (&cplex_env);
		exit (1);
	  }
	  /* Send all log stuff to the cplex.log file. */
	  CPXsetlogfile (cplex_env, fp);

	  /* But discard the results stuff... */
	  CPXdisconnectchannel (cplex_env, _res);
	}
 #else
	/* Older CPLEX library isn't as noisy! */
	_MYCPX_setlogfile (stderr);
 #endif
#endif

	/* Do the branch-and-cut... */
	length = branch_and_cut (&cinfo, &stats, smt_mask);

#if defined (CPLEX) AND (CPLEX >= 40)
	/* Shut down... */
	if (CPXcloseCPLEX (&cplex_env) NE 0) {
		fprintf (stderr, "Warning: Unable to close CPLEX.\n");
	}
#endif

	/* Tally various statistics about the solution.		*/

	edge_count = NEWA (cinfo.num_verts + 1, int);
	total_edge_count = 0;
	smt_edge_count = 0;
	max_edge_size = 0;
	for (i = 0; i <= cinfo.num_verts; i++) {
		edge_count [i] = 0;
	}

	for (i = 0; i < cinfo.num_edges; i++) {
		if (NOT BITON (smt_mask, i)) continue;

		nt = cinfo.edge_size [i];

		++(edge_count [nt]);
		total_edge_count += nt;
		if (nt > max_edge_size) {
			max_edge_size = nt;
		}
		++smt_edge_count;
	}

	if (cinfo.full_trees NE NULL) {
		/* Print out a certificate of the solution.  This	*/
		/* consists of the coordinates of each of the Steiner	*/
		/* points.						*/

		printf ("\n %% Certificate of solution:\n");
		for (i = 0; i < cinfo.num_edges; i++) {
			if (NOT BITON (smt_mask, i)) continue;

			pts = cinfo.full_trees [i] -> steiners;
			for (j = 0; j < pts -> n; j++) {
				coord_to_string (buf1, pts -> a [j].x);
				coord_to_string (buf2, pts -> a [j].y);
				printf (" %% @C	%s	%s\n", buf1, buf2);
			}
		}
	}

	convert_cpu_time (stats.p1time + stats.p2time, buf1);
	dist_to_string (buf2, length);

	descr = "Steiner Minimal Tree";
	if ((cinfo.description NE NULL) AND (cinfo.description [0] NE '\0')) {
		descr = cinfo.description;
	}
	sprintf (title,
		 "%s:  %lu points,  length = %s,  %s seconds",
		 descr, cinfo.num_verts, buf2, buf1);

	overlay_plot_subset (title, smt_mask, &cinfo, BIG_PLOT);

	/* Print out statistics for this run. */
	gap = 100.0 * (stats.z - stats.root_z) / stats.z;
	convert_cpu_time (stats.p1time, buf3);
	convert_cpu_time (stats.p2time, buf4);

	/* Problem summary... */
	printf ("%% @0 %s\n",
		cinfo.description NE NULL ? cinfo.description : "");
	printf ("%% N M Nodes LPs P1CPU P2CPU TotCPU\n");
	printf ("%% @1 %d %d %d %d %s %s %s\n",
		stats.n,
		stats.m,
		stats.num_nodes,
		stats.num_lps,
		buf3, buf4, buf1);

	/* Solution and root node statistics... */
	if (stats.root_opt) {
		sprintf (buf3, "%18.6f", stats.root_z / cinfo.scale_value);
	}
	else {
		sprintf (buf3, "(%18.6f)", stats.root_z / cinfo.scale_value);
	}
	cp1 = &buf3 [0];
	cp2 = &buf3 [0];
	for (;;) {		/* delete spaces... */
		c = *cp2++;
		if ((c NE ' ') AND ((*cp1++ = c) EQ '\0')) break;
	}

	convert_cpu_time (stats.root_time, buf4);
	if (cinfo.mst_length > 0) {
		redmst = 100.0 * (cinfo.mst_length - length)
				/ cinfo.mst_length;
	}
	else {
		redmst = 0.0;
	}
	printf ("%% Z RootZ %%Gap RootLPs RootCPU RedMST\n");
	printf ("%% @2 %s %s %7.5f %d %s %.4f\n",
		buf2,
		buf3,
		gap,
		stats.root_lps,
		buf4,
		redmst);

	/* Initial constraint pool statistics... */
	printf ("%% InitPRows InitPNZ InitLPRows InitLPNZ\n");
	printf ("%% @3 %d %d %d %d\n",
		stats.cs_init.num_prows,
		stats.cs_init.num_pnz,
		stats.cs_init.num_lprows,
		stats.cs_init.num_lpnz);

	/* Root constraint pool statistics... */
	printf ("%% RootPRows RootPNZ RootLPRows RootLPNZ\n");
	printf ("%% @4 %d %d %d %d\n",
		stats.cs_root.num_prows,
		stats.cs_root.num_pnz,
		stats.cs_root.num_lprows,
		stats.cs_root.num_lpnz);

	/* Final constraint statistics... */
	printf ("%% FinalPRows FinalPNZ FinalLPRows FinalLPNZ\n");
	printf ("%% @5 %d %d %d %d\n",
		stats.cs_final.num_prows,
		stats.cs_final.num_pnz,
		stats.cs_final.num_lprows,
		stats.cs_final.num_lpnz);

	/* Statistics on the SMT: number of FSTs, size and distribution. */

	edge10 = 0;
	printf ("%% SMTFSTs SMTAvgFSTSz SMTMaxFSTSz #2FSTs #3FSTs ... #10FSTS #>10FSTs\n");
	printf ("%% @6 %d %f %d",
		smt_edge_count,
		((double) total_edge_count) / ((double) smt_edge_count),
		max_edge_size);
	for (i = 2; i <= 10; i++) {
		j = (i <= cinfo.num_verts) ? edge_count [i] : 0;
		edge10 += j;
		printf (" %d", j);
	}
	printf (" %d\n", smt_edge_count - edge10);

	free ((char *) smt_mask);

	free_phase_1_data (&cinfo);

	return (0);
}

/*
 * This routine decodes the various command-line arguments.
 */

	static
	void
decode_params (

int		argc,
char **		argv
)
{
char *		ap;
char		c;

	--argc;
	me = *argv++;

	printf (" %% %s\n", me);
	printf (" %% Args:\n");

	while (argc > 0) {
		ap = *argv++;
		if (*ap NE '-') {
			usage ();
		}
		printf (" %%	%s\n", ap);
		++ap;
		while ((c = *ap++) NE '\0') {
			switch (c) {
			case '2':
				Seed_Pool_With_2SECs = FALSE;
				break;

			case 'b':
				Choose_Branch_Vars_Carefully = FALSE;
				break;

			case 'l':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					printf (" %%	%s\n", ap);
					--argc;
				}
				if (NOT decode_cpu_time_limit (ap, &cpu_time_limit)) {
					usage ();
				}
				ap = "";
				break;

#ifdef LPSOLVE
			case 'p':
				Use_Perturbations = TRUE;
				break;
#endif

			case 'r':
				Print_Root_LP = TRUE;
				break;

			case 'R':
				Check_Root_Constraints = TRUE;
				break;

#ifdef LPSOLVE
			case 's':
				Use_Scaling = TRUE;
				break;
#endif

			case 't':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					printf (" %%	%s\n", ap);
					--argc;
				}
				Trial_Branch_Threshold = atof (ap);
				ap = "";
				break;

			case 'u':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					printf (" %%	%s\n", ap);
					--argc;
				}
				Initial_Upper_Bound = atof (ap);
				ap = "";
				break;

			default:
				usage ();
				break;
			}
		}
		--argc;
	}
}

/*
 * This routine prints out the proper usage and exits.
 */

static char *	arg_doc [] = {
	"",
	"\t-2\tOmit all 2-terminal SECs from the initial",
	"\t\t constraint pool.",
	"\t-b\tDisable \"strong branching\", which chooses",
	"\t\tbranching variables very carefully.",
	"\t\t (Only needed for old phase 1 data files.)",
#ifdef LPSOLVE
	"\t-p\tUse perturbations when solving LP's.",
#endif
	"\t-r\tPrint root LP relaxation, if fractional.",
	"\t-R\tWhen optimal root LP relaxation is obtained,",
	"\t\tdetermine for each LP iteration the number of",
	"\t\tfinal constraints whose first violation occurred",
	"\t\tduring that iteration.",
#ifdef LPSOLVE
	"\t-s\tUse scaling when solving LP's.",
#endif
	"\t-t T\tStart testing branch variables when",
	"\t\tlower bound meets or exceeds threshold T.",
	"\t-u B\tSets the initial upper bound to B.",
	"",
	"\tExample CPU times are:",
	"\t\t-l 3days2hours30minutes15seconds",
	"\t\t-l1000seconds -l1000 -l 2h30m",
	"",
	NULL
};

	static
	void
usage (void)

{
char **		pp;
char *		p;

	(void) fprintf (stderr,
			"\nUsage: %s"
			" [-2b"
#ifdef LPSOLVE
			"p"
#endif
			"rR"
#ifdef LPSOLVE
			"s"
#endif
			"]"
			" [-l cpu-time-limit] [-m N]"
			" [-t threshold]"
			" [-u upper-bound] <phase-1-data-file\n",
			me);
	pp = &arg_doc [0];
	while ((p = *pp++) NE NULL) {
		(void) fprintf (stderr, "%s\n", p);
	}
	exit (1);
}

/*
 * This routine is the top-level of the branch-and-cut.
 */

	static
	dist_t
branch_and_cut (

struct cinfo *		cip,		/* IN - compatibility info */
struct bbstats *	statp,		/* IN/OUT - statistics */
bitmap_t *		smt		/* OUT - Steiner minimal tree */
)
{
int			i;
int			j;
int			nmasks;
int			nedges;
LP_t *			lp;
int			status;
int			processing_node;
bitmap_t *		vert_mask;
bitmap_t *		edge_mask;
bitmap_t *		req_edges;
bitmap_t *		fixed;
bitmap_t *		value;
bitmap_t *		delta;
double *		x;
double			z;
double			z0;
double			z1;
double			toosmall;
double			toobig;
cpu_time_t		t1;
RETSIGTYPE		(*save_sigterm) (int);
struct cpool *		cpool;
struct cs_info		cs_info;
struct bbtree		bbtree;
struct bbnode *		node;
struct bbnode *		node_to_free;
struct bbnode *		node2;
struct bbnode		root;
struct bbinfo		bbinfo;
struct lpmem		lpmem;

#ifdef CPLEX
int *			b_index;
char *			b_lu;
double *		b_bd;
#endif

	if (Check_Root_Constraints) {
		rcfile = fopen ("/tmp/lp.x", "w");
		if (rcfile EQ NULL) {
			fprintf (stderr,
				 "Warning: unable to create /tmp/lp.x\n");
			Check_Root_Constraints = FALSE;
		}
	}

	bbinfo.t0 = get_cpu_time ();

	/* Establish handler for the SIGTERM (kill -15) signal... */
	force_branch_flag = FALSE;
	save_sigterm = signal (SIGTERM, force_branch_signal_handler);

	nmasks = cip -> num_edge_masks;
	nedges = cip -> num_edges;

	vert_mask	= cip -> initial_vert_mask;
	edge_mask	= cip -> initial_edge_mask;
	req_edges	= cip -> required_edges;

	/* initialize global pool of constraints. */
	cpool = NEW (struct cpool);
	initialize_constraint_pool (cpool, vert_mask, edge_mask, cip);

	/* Build initial formulation. */
	lp = build_initial_formulation (cpool,
					vert_mask,
					edge_mask,
					cip,
					&lpmem);

#if 0
	/* Build cutset separation formulation. */
	build_cutset_separation_formulation (vert_mask,
					     edge_mask,
					     cip,
					     &cs_info);
#endif

	processing_node = 0;

	/* Initialize the branch-and-bound tree... */
	bbtree.first		= NULL;
	bbtree.free		= NULL;
	bbtree.snum		= 0;
	bbtree.nmasks		= nmasks;
	bbtree.node_policy	= NN_BEST_NODE;

	/* Initialize best and worst order heaps */
	bbheap_init (&bbtree.heap [BEST_NODE_HEAP], node_is_better);
	bbheap_init (&bbtree.heap [WORST_NODE_HEAP], node_is_worse);

	/* Create vectors to describe the current problem... */
	fixed	= NEWA (nmasks, bitmap_t);
	value	= NEWA (nmasks, bitmap_t);
	delta	= NEWA (nmasks, bitmap_t);

	for (i = 0; i < nmasks; i++) {
		fixed [i] = 0;
		value [i] = 0;
	}
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (edge_mask, i)) {
			/* variables that are outside of the problem */
			/* are fixed at zero... */
			SETBIT (fixed, i);
			change_var_bounds (lp, i, 0.0, 0.0);
		}
		else if (BITON (req_edges, i)) {
			/* Front-end has determined that this hyperedge	*/
			/* MUST be present in an optimal solution!	*/
			SETBIT (fixed, i);
			SETBIT (value, i);
			change_var_bounds (lp, i, 1.0, 1.0);
		}
	}

	/* Create the root node... */
	root.z		= 0.0;
	root.optimal	= FALSE;
	root.num	= bbtree.snum++;
	root.iter	= 0;
	root.parent	= -1;
	for (i = 0; i < NUM_BB_HEAPS; i++) {
		root.index [i] = -1;
	}
	root.var	= -1;
	root.dir	= 0;
	root.depth	= 0;
	root.br1cnt	= 0;
	root.fixed	= fixed;
	root.value	= value;
	root.n_uids	= 0;
	root.bc_uids	= NULL;
	root.xhist	= NEWA (NUM_X_HISTORIES_PER_NODE * nedges, double);
	root.xh_index	= 0;
	root.next	= NULL;
	root.prev	= NULL;

	for (i = 0; i < NUM_X_HISTORIES_PER_NODE * nedges; i++) {
		root.xhist [i] = 0.0;
	}

#ifdef CPLEX
	/* Create arrays for changing variable bounds... */
	b_index	= NEWA (2 * nedges, int);
	b_lu	= NEWA (2 * nedges, char);
	b_bd	= NEWA (2 * nedges, double);
#endif

	node = &root;

	x	= NEWA (nedges, double);

	/* Fill in the global brand-and-bound info structure... */
	bbinfo.cip		= cip;
	bbinfo.tmap		= vert_mask;
	bbinfo.fset_mask	= edge_mask;
	bbinfo.lp		= lp;
	bbinfo.lpmem		= &lpmem;
	bbinfo.cpool		= cpool;
	bbinfo.bbtree		= &bbtree;
	bbinfo.csip		= &cs_info;
	bbinfo.preempt_z	= Initial_Upper_Bound;
	bbinfo.best_z		= Initial_Upper_Bound;
	bbinfo.smt		= smt;
	bbinfo.node		= node;
	bbinfo.z		= node -> z;
	bbinfo.x		= x;
	bbinfo.slack_size	= 0;
	bbinfo.slack		= NULL;
	bbinfo.dj		= NEWA (nedges, double);
	bbinfo.fixed		= fixed;
	bbinfo.value		= value;
	bbinfo.statp		= statp;
	bbinfo.prevlb		= 0.0;
	bbinfo.fstmst_index	= compute_fstmst_index (cip);

	statp -> n		= cip -> num_verts;
	statp -> m		= nedges;
	statp -> num_nodes	= 0;
	statp -> num_lps	= 0;

	statp -> cs_init.num_prows	= cpool -> nrows;
	statp -> cs_init.num_lprows	= GET_LP_NUM_ROWS (lp);
	statp -> cs_init.num_pnz	= cpool -> num_nz;
	statp -> cs_init.num_lpnz	= GET_LP_NUM_NZ (lp);

	for (;;) {
		/* Set up new node to be processed */
		bbinfo.node = node;
		if (node -> iter <= 0) {
			/* Haven't processed this node before	*/
			/* -- tally another node...		*/
			++(statp -> num_nodes);
		}

		/* Process the current node... */
		status = compute_good_lower_bound (&bbinfo);

		if (node -> depth EQ 0) {
			/* Finished the root node... */
			statp -> root_z = bbinfo.z;
			statp -> root_lps = statp -> num_lps;

			/* Slack rows should have already been deleted... */
			statp -> cs_root.num_prows = cpool -> nrows;
			statp -> cs_root.num_lprows =
				GET_LP_NUM_ROWS (bbinfo.lp);
			statp -> cs_root.num_pnz = cpool -> num_nz;
			statp -> cs_root.num_lpnz =
				GET_LP_NUM_NZ (bbinfo.lp);
			statp -> root_opt = node -> optimal;

			t1 = get_cpu_time ();
			statp -> root_time = t1 - bbinfo.t0;

			if ((status EQ LB_FRACTIONAL) AND Print_Root_LP) {
				print_root_lp (&bbinfo);
			}

			if (Check_Root_Constraints) {
				check_root_constraints (&bbinfo);
			}
		}

		node_to_free = NULL;	/* Default is no node to free... */

		switch (status) {
		case LB_INFEASIBLE:
			/* Node is fathomed! */
			trace_node (&bbinfo, ' ', "infeasible");
			node_to_free = node;
			break;

		case LB_CUTOFF:
			/* Node is fathomed! */
			trace_node (&bbinfo, ' ', "cutoff");
			node_to_free = node;
			break;

		case LB_INTEGRAL:
			if (bbinfo.z >= bbinfo.best_z) {
				trace_node (&bbinfo, ' ', "cutoff");
				break;
			}
			/* We have a new best feasible integer solution! */
			for (i = 0; i < nmasks; i++) {
				smt [i] = 0;
			}
			for (i = 0; i < nedges; i++) {
				if (x [i] >= 0.5) {
					SETBIT (smt, i);
				}
			}
			new_upper_bound (bbinfo.z, &bbinfo);

			trace_node (&bbinfo, '*', NULL);

			node_to_free = node;
			break;

		case LB_FRACTIONAL:
			j = choose_branching_variable (&bbinfo, &z0, &z1);

			/* Create two nodes... */
			if (z0 < z1) {
				add_bbnode (&bbinfo, j, 0, z0);
				add_bbnode (&bbinfo, j, 1, z1);
			}
			else if (z1 < z0) {
				add_bbnode (&bbinfo, j, 1, z1);
				add_bbnode (&bbinfo, j, 0, z0);
			}
			else if (UP_FIRST) {	/* To break ties... */
				add_bbnode (&bbinfo, j, 0, z0);
				add_bbnode (&bbinfo, j, 1, z1);
			}
			else {
				add_bbnode (&bbinfo, j, 1, z1);
				add_bbnode (&bbinfo, j, 0, z0);
			}
			trace_node (&bbinfo, ' ', NULL);

			/* This node is done (became 2 children), free it. */
			node_to_free = node;
			break;

		case LB_PREEMPTED:
			printf ("%% suspending node %d\n", node -> num);
			/* This node is no longer the best.  Put it	*/
			/* back into the heap and get another one...	*/
			node2 = bbtree.first;
			if (node2 NE NULL) {
				node2 -> prev = node;
			}
			node -> next = node2;
			node -> prev = NULL;
			bbtree.first = node;

			bbheap_insert (node, &bbtree, BEST_NODE_HEAP);
			bbheap_insert (node, &bbtree, WORST_NODE_HEAP);

			/* Deactivating this node -- remember the basis */
			save_node_basis (node, &bbinfo);

			/* Do NOT free this node! */
			break;
		}

		/* If there is a node to free, do so now... */
		if ((node_to_free NE NULL) AND (node_to_free NE &root)) {
			/* Free up saved basis info and decrement	*/
			/* constraint reference counts before freeing.	*/
			destroy_node_basis (node_to_free, &bbinfo);

			node_to_free -> next = bbtree.free;
			bbtree.free = node_to_free;
		}

		/* Select the next node to process. */
		node = select_next_node (&bbtree);
		if (node EQ NULL) break;

		/* This is perhaps a new lower bound... */
		new_lower_bound (node -> z, &bbinfo);

		printf ("%% Resuming node %d\n", node -> num);

		/* Restore the LP tableaux and basis for this node.	*/
		/* Decrement the reference counts on this node's	*/
		/* binding rows.  Since it is now the ACTIVE node,	*/
		/* there is no reason to continue protecting these rows	*/
		/* from deletion by other nodes.			*/
		restore_node_basis (node, &bbinfo);

		/* Determine new preemption value (i.e. the objective	*/
		/* value of the next-best node).			*/
		update_node_preempt_value (&bbinfo);

		/* Modify LP to represent problem from new node. */
		for (i = 0; i < nmasks; i++) {
			delta [i] =   (fixed [i] ^ node -> fixed [i])
				    | (value [i] ^ node -> value [i]);
		}
#ifdef CPLEX
		j = 0;
		for (i = 0; i < nedges; i++) {
			if (NOT BITON (delta, i)) continue;
			/* Force bounds for variable 'i' to be correct... */
			b_index [j]	= i;	/* variable i, */
			b_lu [j]	= 'L';	/*	lower bound */
			b_index [j+1]	= i;	/* variable i, */
			b_lu [j+1]	= 'U';	/*	upper bound */
			if (NOT BITON (node -> fixed, i)) {
				/* new variable is NOT fixed... */
				b_bd [j]	= 0.0;
				b_bd [j+1]	= 1.0;
			}
			else if (NOT BITON (node -> value, i)) {
				/* new variable is fixed to 0 */
				b_bd [j]	= 0.0;
				b_bd [j+1]	= 0.0;
			}
			else {
				/* new variable is fixed to 1 */
				b_bd [j]	= 1.0;
				b_bd [j+1]	= 1.0;
			}
			j += 2;
		}
		if (_MYCPX_chgbds (bbinfo.lp, j, b_index, b_lu, b_bd) NE 0) {
			fatal ("branch_and_cut: Bug 1.");
		}
#endif

#ifdef LPSOLVE
		for (i = 0; i < nedges; i++) {
			if (NOT BITON (delta, i)) continue;
			/* Force bounds on variable 'i' to be correct... */
			if (NOT BITON (node -> fixed, i)) {
				/* variable is NOT fixed... */
				set_bounds (lp, i + 1, 0.0, 1.0);
			}
			else if (NOT BITON (node -> value, i)) {
				/* variable is fixed to 0 */
				set_bounds (lp, i + 1, 0.0, 0.0);
			}
			else {
				/* variable is fixed to 1 -- must set	*/
				/* bounds in this order to avoid	*/
				/* lb > ub condition between calls...	*/
				set_bounds (lp, i + 1, 1.0, 1.0);
			}
		}
#endif

		for (i = 0; i < nmasks; i++) {
			fixed [i] = node -> fixed [i];
			value [i] = node -> value [i];
		}
	}

	statp -> cs_final.num_prows	= cpool -> nrows;
	statp -> cs_final.num_lprows	= GET_LP_NUM_ROWS (bbinfo.lp);
	statp -> cs_final.num_pnz	= cpool -> num_nz;
	statp -> cs_final.num_lpnz	= GET_LP_NUM_NZ (bbinfo.lp);

	/* New lower bound! */
	new_lower_bound (bbinfo.best_z, &bbinfo);

	t1 = get_cpu_time ();
	statp -> p2time = t1 - bbinfo.t0;
	statp -> z = bbinfo.best_z;

	/* Restore normal handling of SIGTERM... */
	(void) signal (SIGTERM, save_sigterm);

	destroy_initial_formulation (&bbinfo);

	free ((char *) bbinfo.fstmst_index);
	free ((char *) bbinfo.dj);
	if (bbinfo.slack NE NULL) {
		free ((char *) bbinfo.slack);
	}
	free ((char *) x);
#if CPLEX
	free ((char *) b_bd);
	free ((char *) b_lu);
	free ((char *) b_index);
#endif

	free ((char *) root.xhist);

	free ((char *) delta);
	free ((char *) value);
	free ((char *) fixed);

	bbheap_free (&bbtree.heap [BEST_NODE_HEAP]);
	bbheap_free (&bbtree.heap [WORST_NODE_HEAP]);

	free_constraint_pool (cpool);

	/* Return length of final tree... */

	return ((dist_t) bbinfo.best_z);
}

/*
 * A handler for the SIGTERM signal.  When we receive this signal we
 * stop generating constraints for the current node and branch instead.
 */

	static
	RETSIGTYPE
force_branch_signal_handler (

int		sig		/* IN - signal being handled */
)
{
	/* Stop generating constraints and force a branch instead. */
	force_branch_flag = TRUE;

	/* Prepare to handle the signal again... */
	(void) signal (SIGTERM, force_branch_signal_handler);
}

/*
 * This routine adds a new node to the branch-and-bound tree.
 */

	static
	void
add_bbnode (

struct bbinfo *		bbip,		/* IN - branch-and-bound info */
int			var,		/* IN - variable to branch on */
int			dir,		/* IN - branch direction */
double			z		/* IN - value to give to node */
)
{
int			i;
int			j;
int			n;
int			nmasks;
struct bbnode *		parent;
struct bbtree *		tp;
struct bbnode *		p;
struct bbnode *		p1;
struct bbnode *		p2;
struct bbnode **	tmp1;
struct bbnode **	tmp2;

	if (z >= bbip -> best_z) {
		/* Node is already worse than the cutoff value!	*/
		/* Don't bother adding it to the tree!		*/
		return;
	}

	parent	= bbip -> node;		/* current node becomes parent */
	tp	= bbip -> bbtree;	/* tree to insert in */

	printf (" %% @NC %4d %4d	x%d = %d	%f\n",
		tp -> snum, parent -> num, var, dir, z);

	nmasks	= tp -> nmasks;

	n = NUM_X_HISTORIES_PER_NODE * bbip -> cip -> num_edges;

	/* Get a new tree node... */
	p = tp -> free;
	if (p NE NULL) {
		tp -> free = p -> next;
	}
	else {
		p = NEW (struct bbnode);
		p -> fixed = NEWA (2 * nmasks, bitmap_t);
		p -> value = p -> fixed + nmasks;
		p -> xhist = NEWA (n, double);
	}
	p -> z		= z;
	p -> optimal	= FALSE;
	p -> num	= (tp -> snum)++;
	p -> iter	= 0;
	p -> parent	= parent -> num;
	p -> var	= var;
	p -> dir	= dir;
	p -> depth	= 1 + parent -> depth;
	if (dir EQ 0) {
		p -> br1cnt = parent -> br1cnt;
	}
	else {
		p -> br1cnt = parent -> br1cnt + 1;
	}

	/* Get most up-to-date fixed variables... */
	for (i = 0; i < nmasks; i++) {
		p -> fixed [i] = bbip -> fixed [i];
		p -> value [i] = bbip -> value [i];
	}
	SETBIT (p -> fixed, var);
	if (dir EQ 0) {
		CLRBIT (p -> value, var);
	}
	else {
		SETBIT (p -> value, var);
	}
	p -> n_uids	= 0;
	p -> bc_uids	= NULL;
	p -> bc_row	= NULL;
	p -> rstat	= NULL;
	p -> cstat	= NULL;

	/* Copy parent's LP solution history... */
	memcpy (p -> xhist, parent -> xhist, n * sizeof (p -> xhist [0]));
	p -> xh_index = parent -> xh_index;

	/* Save the current basis (actually the parent's basis)	*/
	/* into this node.					*/
	save_node_basis (p, bbip);

	/* Insert node into depth-first list... */
	p1 = tp -> first;
	if (p1 NE NULL) {
		p1 -> prev = p;
	}
	p -> next	= p1;
	p -> prev	= NULL;
	tp -> first = p;

	/* Insert node into "best-node" heap... */
	bbheap_insert (p, tp, BEST_NODE_HEAP);

	/* Insert node into "worst-node" heap... */
	bbheap_insert (p, tp, WORST_NODE_HEAP);
}

/*
 * This routine selects the next node to process from the given
 * branch-and-bound tree.  This is where we implement the specific
 * search policy.
 */

	static
	struct bbnode *
select_next_node (

struct bbtree *		tp	/* IN - branch-and-bound tree */
)
{
struct bbnode *		p;

	if (tp -> first EQ NULL) {
		/* No more nodes! */
		return (NULL);
	}

	switch (tp -> node_policy) {
	case NN_DEPTH_FIRST:
		/* Get node created most recently... */
		p = tp -> first;
		break;

	case NN_BEST_NODE:
		/* Get node with lowest objective function value... */
		p = tp -> heap [BEST_NODE_HEAP].array [0];
		break;

	default:
		fatal ("select_next_node: Bug 1.");
	}

	delete_node_from_bbtree (p, tp);

	return (p);
}

/*
 * This routine deletes an arbitrary node from an arbitrary position
 * within the given branch-and-bound tree.
 */

	static
	void
delete_node_from_bbtree (

struct bbnode *		p,	/* IN - node to delete */
struct bbtree *		tp	/* IN - tree to delete it from */
)
{
struct bbnode * p1;
struct bbnode * p2;


	/* Delete it from LIFO list... */
	p1 = p -> prev;
	p2 = p -> next;
	if (p1 NE NULL) {
		p1 -> next = p2;
	}
	else {
		tp -> first = p2;
	}
	if (p2 NE NULL) {
		p2 -> prev = p1;
	}
	p -> next = NULL;
	p -> prev = NULL;

	/* Delete it from the best-node heap... */
	bbheap_delete (p, tp, BEST_NODE_HEAP);

	/* Delete it from the worst-node heap... */
	bbheap_delete (p, tp, WORST_NODE_HEAP);
}

/*
 * This routine initializes a branch-and-bound node heap.
 */

	static
	void
bbheap_init (

struct bbheap *		hp,	/* IN - the heap to initialize */
bbheap_func_t *		funcp	/* IN - node comparison function */
)
{
	hp -> array		= NEWA (INIT_HSIZE, struct bbnode *);
	hp -> hsize		= INIT_HSIZE;
	hp -> nheap		= 0;
	hp -> is_parent_funcp	= funcp;
}


	static
	void
bbheap_free (

struct bbheap *		hp	/* IN - heap to free up */
)
{
	free ((char *) (hp -> array));
}

/*
 * This routine adds the given branch-and-bound node to the specified
 * heap of the given branch-and-bound tree.
 */

	static
	void
bbheap_insert (

struct bbnode *		p,	/* IN - node to insert */
struct bbtree *		tp,	/* IN - branch-and-bound tree with heaps */
int			heap_no	/* IN - heap number to insert node into */
)
{
int			i;
int			j;
int			n;
struct bbheap *		hp;
struct bbnode **	tmp;
struct bbnode *		p2;
bbheap_func_t *		is_parent;

	/* Use specified heap... */
	hp = &(tp -> heap [heap_no]);

	/* Verify sufficient heap space to hold new node... */
	n = hp -> nheap;
	if (n >= hp -> hsize) {
		tmp = NEWA (2 * n, struct bbnode *);
		for (i = 0; i < n; i++) {
			tmp [i] = hp -> array [i];
		}
		free ((char *) (hp -> array));
		hp -> array = tmp;
		hp -> hsize = 2 * n;
	}

	is_parent = hp -> is_parent_funcp;

	/* Insert node into heap by placing it at the end of	*/
	/* the array and sifting up...				*/
	for (i = n; i > 0;) {
		j = ((i - 1) >> 1);
		p2 = hp -> array [j];
		if (is_parent (p2, p)) break;
		p2 -> index [heap_no] = i;
		hp -> array [i] = p2;
		i = j;
	}
	p -> index [heap_no] = i;
	hp -> array [i] = p;
	hp -> nheap = n + 1;
}

/*
 * This routine deletes a single node from the specified heap of the
 * given branch-and-bound tree.
 */

	static
	void
bbheap_delete (

struct bbnode *		p,	/* IN - the node to delete from the heap */
struct bbtree *		tp,	/* IN - branch-and-bound tree with heaps */
int			heap_no	/* IN - heap number from which to delete */
)
{
int			i;
int			j;
int			n;
struct bbheap *		hp;
struct bbnode *		p1;
struct bbnode *		p2;
struct bbnode *		p3;
bbheap_func_t *		is_parent;

	/* Use proper heap to delete from... */
	hp = &(tp -> heap [heap_no]);
	n = hp -> nheap;

	if ((n <= 0) OR (n > hp -> hsize)) {
		fatal ("delete_bbheap: Bug 1.");
	}

	/* Deleting from a heap requires three steps:		*/
	/*	1. Move the last node into the deleted node's	*/
	/*	   current position.				*/
	/*	2. Sift this replacement node up.		*/
	/*	3. Sift this replacement node down.		*/

	/* Get position of node being deleted... */
	i = p -> index [heap_no];
	if (i >= n) {
		fatal ("delete_bbheap: Bug 2.");
	}
	if (hp -> array [i] NE p) {
		fatal ("delete_bbheap: Bug 3.");
	}

	/* Deleted node is no longer in the array... */
	p -> index [heap_no] = -1;

	/* Heap now has one fewer element in it... */
	--n;
	hp -> nheap = n;
	if (n <= 0) {
		/* Heap is now empty -- done! */
		return;
	}

	/* Move last heap element into vacated spot. */
	p1 = hp -> array [n];
	if (p1 EQ p) {
		/* Deleting last item from heap -- we are done! */
		return;
	}
	if (p1 -> index [heap_no] NE n) {
		fatal ("delete_bbheap: Bug 4.");
	}

	/* We now assume that node "p1" will be in position "i"... */

	is_parent = hp -> is_parent_funcp;

	/* First sift up... */
	while (i > 0) {
		j = ((i - 1) >> 1);
		p2 = hp -> array [j];
		if (is_parent (p2, p1)) break;
		p2 -> index [heap_no] = i;
		hp -> array [i] = p2;
		i = j;
	}

	/* Now sift down... */
	while (i < n) {
		j = (i << 1) + 1;
		if (j >= n) break;
		p2 = hp -> array [j];
		if ((j + 1) < n) {
			p3 = hp -> array [j + 1];
			if (is_parent (p3, p2)) {
				++j;
				p2 = p3;
			}
		}
		if (is_parent (p1, p2)) break;
		p2 -> index [heap_no] = i;
		hp -> array [i] = p2;
		i = j;
	}
	p1 -> index [heap_no] = i;
	hp -> array [i] = p1;

	hp -> nheap = n;
}

/*
 * This routine returns TRUE if-and-only-if node 1 is "better" than
 * node 2 -- in other words, node 1 must be above node 2 in the
 * "best-node" heap.
 */

	static
	int
node_is_better (

struct bbnode *		p1,	/* IN - node 1 */
struct bbnode *		p2	/* IN - node 2 */
)
{
	if (p1 -> z < p2 -> z) return (TRUE);
	if (p1 -> z > p2 -> z) return (FALSE);

	/* Objective values are equal -- use node creation	*/
	/* order to break the tie.  What we want to achieve is	*/
	/* that the "var=0" and "var=1" children of a node	*/
	/* (which will have equal objective values) get taken	*/
	/* from the "best-node" heap in proper order with	*/
	/* respect to the UP_FIRST switch.  (The node that was	*/
	/* created last should be taken from the heap first.)	*/

	if (p1 -> num >= p2 -> num) return (TRUE);
	return (FALSE);
}


/*
 * This routine returns TRUE if-and-only-if node 1 is "worse" than
 * node 2 -- in other words, node 2 must be above node 2 in the
 * "worst-node" heap.
 */

	static
	int
node_is_worse (

struct bbnode *		p1,	/* IN - node 1 */
struct bbnode *		p2	/* IN - node 2 */
)
{
	return (p1 -> z >= p2 -> z);
}

/*
 * This routine traces the result for a given node.
 */

	static
	void
trace_node (

struct bbinfo *		bbip,	/* IN - the branch-and-bound info */
char			c1,	/* IN - space or * char */
char *			msg	/* IN - message OR NULL */
)
{
struct bbnode *		p;
struct bbtree *		tp;
struct bbheap *		hp;
double			lowz;
char *			p1;
char *			p2;
char			c2;
char			buf1 [32];
char			buf2 [32];
char			buf3 [32];
char			buf4 [32];
char			buf5 [32];
char			buf6 [32];
char			line [136];

	p	= bbip -> node;
	tp	= bbip -> bbtree;

	if (msg EQ NULL) {
		(void) sprintf (buf1, "%14.4f", bbip -> z);
	}
	else {
		(void) sprintf (buf1, "%14s", msg);
	}
	if (bbip -> best_z EQ DBL_MAX) {
		buf2 [0] = '\0';
	}
	else {
		(void) sprintf (buf2, "%14.4f", bbip -> best_z);
	}
	hp = &(tp -> heap [BEST_NODE_HEAP]);
	if (hp -> nheap > 0) {
		(void) sprintf (buf3,
				"%14.4f",
				hp -> array [0] -> z);
	}
	else {
		buf3 [0] = '\0';
	}
	if (p -> var < 0) {
		buf4 [0] = '\0';
		c2 = ' ';
	}
	else {
		(void) sprintf (buf4, "x%d", p -> var);
		c2 = (p -> dir EQ 0) ? 'D' : 'U';
	}
	if (p -> parent < 0) {
		buf5 [0] = '\0';
	}
	else {
		(void) sprintf (buf5, "%d", p -> parent);
	}
	if (p -> depth <= 0) {
		buf6 [0] = '\0';
	}
	else {
		(void) sprintf (buf6, "%d", p -> depth);
	}

	(void) sprintf (line,
			"%c%6d%6d%14s%14s%14s%6s %c%6s%6s",
			c1,		/* space or * */
			p -> num,	/* node number */
			hp -> nheap,	/* nodes left */
			buf1,		/* objective/cutoff/infeas */
			buf2,		/* best integer soln */
			buf3,		/* best node */
			buf4,		/* variable name */
			c2,		/* branch direction */
			buf5,		/* parent node number */
			buf6);		/* node depth */

	p1 = line;
	for (p2 = p1; *p2 NE '\0'; p2++) {
	}
	while ((p2 > p1) AND (p2 [-1] EQ ' ')) {
		--p2;
	}
	*p2 = '\0';
	(void) printf ("  %% %s\n", line);
}

/*
 * This routine plots the LP relaxation solution for the root node.  We
 * do this whenever -r is specified and we get an optimal LP relaxation
 * solution that is fractional.
 */

	static
	void
print_root_lp (

struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
struct cinfo *		cip;
struct bbnode *		nodep;
struct bbstats *	statp;
char *			descr;
char			buf1 [32];
char			buf2 [32];
char			title [80];

	cip	= bbip -> cip;
	nodep	= bbip -> node;
	statp	= bbip -> statp;

	convert_cpu_time (statp -> root_time, buf1);
	dist_to_string (buf2, nodep -> z);

	descr = "Steiner Minimal Tree";
	if ((cip -> description NE NULL) AND
	    (cip -> description [0] NE '\0')) {
		descr = cip -> description;
	}
	sprintf (title,
		 "%s:  %lu points,  Root Node, LP %d,  length = %s,  %s seconds",
		 descr,
		 cip -> num_verts,
		 nodep -> iter,
		 buf2,
		 buf1);

	plot_lp_solution (title, bbip -> x, cip, BIG_PLOT);
}

/*
 * This routine chooses the next variable to branch on at this
 * node.
 */

	static
	int
choose_branching_variable (

struct bbinfo *		bbip,		/* IN - branch-and-bound info */
double *		z0,		/* OUT - value to give Xi=0 node */
double *		z1		/* OUT - value to give Xi=1 node */
)
{
int			i;
int			nedges;
int			best_var;
struct cinfo *		cip;
double *		x;
bitmap_t *		fset_mask;
double			xi;
double			max_infeas;
double			infeas;

	if (Choose_Branch_Vars_Carefully) {
		/* Do it very carefully! */
		return (carefully_choose_branching_variable (bbip, z0, z1));
	}

	/* There are LOTS of things that we MIGHT do here in	*/
	/* the FUTURE!  For right now, just take the variable	*/
	/* that is closest to 1/2 -- take larger variables in	*/
	/* the event of a tie...				*/

	cip		= bbip -> cip;
	x		= bbip -> x;
	fset_mask	= bbip -> fset_mask;

	nedges	= cip -> num_edges;

	best_var	= -1;
	max_infeas	= 0.0;

	for (i = nedges - 1; i >= 0; i--) {
		if (NOT BITON (fset_mask, i)) continue;
		xi = x [i];
		if (xi <= FUZZ) continue;
		if (xi + FUZZ >= 1.0) continue;
		infeas = 1.0 - xi;
		if (xi < infeas) {
			infeas = xi;
		}
		if (infeas > max_infeas) {
			best_var = i;
			max_infeas = infeas;
		}
	}
	if (best_var < 0) {
		fatal ("choose_branching_variable: Bug 1.");
	}

	/* Give both nodes the same value... */
	*z0	= bbip -> z;
	*z1	= bbip -> z;

	return (best_var);
}

/*
 * This routine does a very careful job of choosing the next variable
 * to branch on.  For each fractional variable Xi, we solve the LP
 * first with Xi=0 and then Xi=1, yielding objective values Zi0 and Zi1
 * correspondingly.  We then choose variable Xi for which the value
 * min(Zi0, Zi1) is MAXIMIZED.  This provides us with the best possible
 * "one-branch/no-cuts" improvement in the lower bound.
 */

	static
	int
carefully_choose_branching_variable (

struct bbinfo *		bbip,		/* IN - branch-and-bound info */
double *		node_z0,	/* OUT - value for Xi=0 node */
double *		node_z1		/* OUT - value for Xi=1 node */
)
{
int			i;
int			nedges;
int			nfrac;
int			best_var;
double			best_z0;
double			best_z1;
double			best_z;
struct cinfo *		cip;
double *		x;
double *		newx;
bitmap_t *		edge_mask;
bool			found;
LP_t *			lp;
double			xi;
double			z0;
double			z1;
double			z;
struct basis_save	bsave;

	cip		= bbip -> cip;
	lp		= bbip -> lp;
	x		= bbip -> x;
	edge_mask	= bbip -> fset_mask;

	nedges	= cip -> num_edges;

#if 1
	nfrac = 0;
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (edge_mask, i)) continue;
		xi = x [i];
		if (xi <= FUZZ) continue;
		if (xi + FUZZ >= 1.0) continue;
		++nfrac;
	}
	printf ("\n %% Carefully choosing branching variable, nfrac = %d\n",
		nfrac);
#endif

#if 0
	/* Debug code to dump the LP instance we are attempting... */
	{ char		buf [64];
		sprintf (buf,
			 "Try_Branch.%03d.lp",
			 bbip -> node -> num);
		dump_lp (lp, buf);
	}
#endif

	/* Get temporary solution vector... */
	newx = NEWA (nedges, double);

	/* Snapshot the current basis so that we can quickly	*/
	/* get back to it each time...				*/
	save_LP_basis (lp, &bsave);

	best_var	= -1;
	best_z		= 0.0;
	best_z0		= 0.0;
	best_z1		= 0.0;

	for (i = 0; i < nedges; i++) {
		if (NOT BITON (edge_mask, i)) continue;
		xi = x [i];
		if (xi <= FUZZ) continue;
		if (xi + FUZZ >= 1.0) continue;

		/* Solve the Xi=0 case... */
		z0 = try_branch (lp, i + 1, 0, newx, INF_DISTANCE, &bsave);

		/* Check for a better integer feasible solution... */
		found = check_for_better_IFS (newx, bbip, &z0);
#if 1
		printf (" %%%s\tx%d = 0,\tZ0 = %-24.15g\n",
			found ? " !!!" : "", i, z0);
#endif

		if (z0 > best_z) {
			/* Solve the Xi=1 case... */
			z1 = try_branch (lp,
					 i + 1,
					 1,
					 newx,
					 INF_DISTANCE,
					 &bsave);

			/* Check for a better integer feasible solution... */
			found = check_for_better_IFS (newx, bbip, &z1);

#if 1
			printf (" %%%s\tx%d = 1,\tZ1 = %-24.15g\n",
				found ? " !!!" : "", i, z1);
#endif

			z = z0;
			if (z1 < z) {
				z = z1;
			}
			if (z > best_z) {
				best_z		= z;
				best_z0		= z0;
				best_z1		= z1;
				best_var	= i;
#if 1
				printf (" %%   New best:  x%d, Z = %-24.15g\n",
					best_var, best_z);
#endif
				if (best_z >= bbip -> best_z) {
					/* Nice deal!  This variable	*/
					/* forces a cutoff on both	*/
					/* branches!  No need to look	*/
					/* at the rest...		*/
					break;
				}
			}
		}
	}

	if (best_var < 0) {
		fatal ("carefully_choose_branching_variable: Bug 1.");
	}

#if 1
	printf (" %% Best branch is x%d, Z0 = %-24.15g, Z1 = %-24.15g\n\n",
		best_var, best_z0, best_z1);
#endif

	destroy_LP_basis (&bsave);

	free ((char *) newx);

	*node_z0 = best_z0;
	*node_z1 = best_z1;

	return (best_var);
}

/*
 * This routine goes through each of the fractional variables, testing
 * how effectively each one performs as a branching variable.  It
 * prints out a painful amount of debugging output.  The hope is that
 * we can use this output to devise a decent heuristic for choosing
 * variables that are likely to perform well as branching variables.
 */

	static
	void
test_branching_variables (

struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
int			i;
int			j;
int			nedges;
int			nfrac;
int			best_var;
double			best_z0;
double			best_z1;
double			best_z;
struct cinfo *		cip;
double *		x;
double *		newx;
bitmap_t *		tmap;
bitmap_t *		fset_mask;
bool			found0;
bool			found1;
LP_t *			lp;
double			xi;
double			z0;
double			z1;
double			z0_ppm;
double			z1_ppm;
double			zmin;
double			zmax;
double			zmin_ppm;
double			zmax_ppm;
double *		xhist;
double *		xp;
double			sum_xi;
double			sum_xi2;
double			mu_x;
double			var_x;
struct basis_save	bsave;

	cip		= bbip -> cip;
	lp		= bbip -> lp;
	x		= bbip -> x;
	tmap		= bbip -> tmap;
	fset_mask	= bbip -> fset_mask;
	xhist		= bbip -> node -> xhist;

	nedges	= cip -> num_edges;

	nfrac = 0;
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		xi = x [i];
		if (xi <= FUZZ) continue;
		if (xi + FUZZ >= 1.0) continue;
		++nfrac;
	}
	printf ("\n %% Testing all possible branching variables, nfrac = %d\n",
		nfrac);

	/* Get temporary solution vector... */
	newx = NEWA (nedges, double);

	/* Snapshot the current basis so that we can quickly	*/
	/* get back to it each time...				*/
	save_LP_basis (lp, &bsave);

	best_var	= -1;
	best_z		= 0.0;
	best_z0		= 0.0;
	best_z1		= 0.0;

	printf (" %% ..    i   Xi   Infeas Rank N     Z0     Z1    Min    Max\n");
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		xi = x [i];
		if (xi <= FUZZ) continue;
		if (xi + FUZZ >= 1.0) continue;

		/* Solve the Xi=0 case... */
		z0 = try_branch (lp, i + 1, 0, newx, INF_DISTANCE, &bsave);

		z0_ppm = 1000000.0 * ((z0 - bbip -> z) / bbip -> z);

		/* Solve the Xi=1 case... */
		z1 = try_branch (lp, i + 1, 1, newx, INF_DISTANCE, &bsave);

		z1_ppm = 1000000.0 * ((z1 - bbip -> z) / bbip -> z);

		if (z0 < z1) {
			zmin = z0;
			zmax = z1;
			zmin_ppm = z0_ppm;
			zmax_ppm = z1_ppm;
		}
		else {
			zmin = z1;
			zmax = z0;
			zmin_ppm = z1_ppm;
			zmax_ppm = z0_ppm;
		}
		if (zmin > best_z) {
			best_z		= zmin;
			best_z0		= z0;
			best_z1		= z1;
			best_var	= i;
		}

		/* Compute mean and variance of Xi over the last few	*/
		/* iterations.						*/
		sum_xi = 0.0;
		sum_xi2 = 0.0;
		xp = &xhist [i];
		for (j = 0; j < NUM_X_HISTORIES_PER_NODE; j++) {
			sum_xi += *xp;
			sum_xi2 += (*xp * *xp);
			xp += nedges;
		}
		mu_x = sum_xi / NUM_X_HISTORIES_PER_NODE;
		var_x = (sum_xi2 - (sum_xi * sum_xi) / NUM_X_HISTORIES_PER_NODE)
			/ (NUM_X_HISTORIES_PER_NODE - 1);

		printf (" %% @q %4d %6f %6f %4d %2d %6f %6f %6f %6f %6f %6f\n",
			i, xi, 0.5 - fabs (xi - 0.5),
			bbip -> fstmst_index [i],
			cip -> edge_size [i],
			mu_x, var_x,
			z0, z1, zmin_ppm, zmax_ppm);
	}

#if 1
	printf (" %% Best branch is x%d, Z0 = %-24.15g, Z1 = %-24.15g\n\n",
		best_var, best_z0, best_z1);
#endif

	destroy_LP_basis (&bsave);

	free ((char *) newx);
}

/*
 * This routine checks to see if the result of doig a "test-branch" on
 * a variable JUST HAPPENS to result in an integer feasible solution
 * that is the best so far.  Although this would be total serendipity,
 * we do it anyway.  The check takes virtually no time because we almost
 * always locate a fractional variable within the first few probes.
 * And besides, it would be a shame to NOT notice something this important
 * -- especially if we don't have ANY upper bound yet!
 */

	static
	bool
check_for_better_IFS (

double *		x,		/* IN - solution to test */
struct bbinfo *		bbip,		/* IN - branch and bound info */
double *		true_z		/* OUT - true Z value, if integral */
)
{
int			i;
int			nedges;
int			nmasks;
struct cinfo *		cip;
double			z;
int			num_frac;

	cip = bbip -> cip;

	if (NOT integer_feasible_solution (x,
					   bbip -> tmap,
					   bbip -> fset_mask,
					   cip,
					   &num_frac)) {
		/* Not integer feasible -- get out. */
		return (FALSE);
	}

	/* We literally stumbled across an Integer Feasible Solution!	*/

	/* Re-calculate the final objective function (in order to	*/
	/* eliminate some of the LP solver's numerical errors)...	*/

	nedges	= cip -> num_edges;
	z = 0.0;
	for (i = 0; i < nedges; i++) {
		if (x [i] + FUZZ < 1.0) continue;
		z += cip -> cost [i];
	}

	/* Give caller the correct Z value... */
	*true_z = z;

	/* Damn Intel FPU is keeping 80 bits for Z... */
	z += 1.0;
	z = *true_z;

	if (z >= bbip -> best_z) {
		/* No better than what we have... */
		return (FALSE);
	}

	/* We have a new best integer feasible solution!  Record it!	*/
	nmasks = cip -> num_edge_masks;
	for (i = 0; i < nmasks; i++) {
		bbip -> smt [i] = 0;
	}
	for (i = 0; i < nedges; i++) {
		if (x [i] >= 0.5) {
			SETBIT (bbip -> smt, i);
		}
	}

	new_upper_bound (z, bbip);

	return (TRUE);
}

/*
 * This routine saves the current basis of the given LP.
 */

#ifdef CPLEX

	static
	void
save_LP_basis (

LP_t *			lp,		/* IN - LP to save basis for */
struct basis_save *	basp		/* OUT - saved basis info */
)
{
int		rows;
int		cols;

	rows = _MYCPX_getnumrows (lp);
	cols = _MYCPX_getnumcols (lp);

	basp -> cstat = NEWA (cols, int);
	basp -> rstat = NEWA (rows, int);

	if (_MYCPX_getbase (lp, basp -> cstat, basp -> rstat) NE 0) {
		fatal ("save_LP_basis: Bug 1.");
	}
}

/*
 * Destroy the saved basis info...
 */

	static
	void
destroy_LP_basis (

struct basis_save *	basp		/* IN - basis info to free up */
)
{
#if CPLEX >= 40
	free ((char *) (basp -> rstat));
	free ((char *) (basp -> cstat));
#endif
}

#endif

/*
 * This routine tries the given branch by solving the LP.  It
 * returns the resulting objective value, or INF_DISTANCE if something
 * goes wrong (like infeasible).
 */

#ifdef CPLEX

	static
	double
try_branch (

LP_t *			lp,		/* IN - LP to re-optimize */
int			var,		/* IN - variable to try branching */
int			dir,		/* IN - branch direction, 0 or 1 */
double *		x,		/* OUT - LP solution obtained */
double			ival,		/* IN - value to give if infeasible */
struct basis_save *	basp		/* IN - basis to restore when done */
)
{
int		status;
double		z;
int		b_index [2];
char		b_lu [2];
double		b_bd [2];

	--var;		/* vars are zero-origined in CPLEX... */

	b_index [0] = var;	b_lu [0] = 'L';
	b_index [1] = var;	b_lu [1] = 'U';
	if (dir EQ 0) {
		b_bd [0] = 0.0;
		b_bd [1] = 0.0;
	}
	else {
		b_bd [0] = 1.0;
		b_bd [1] = 1.0;
	}
	if (_MYCPX_chgbds (lp, 2, b_index, b_lu, b_bd) NE 0) {
		fatal ("try_branch: Bug 1.");
	}

	/* Solve the current LP instance... */
	status = _MYCPX_dualopt (lp);
	if (status NE 0) {
		printf (" %%  WARNING dualopt: status = %d\n", status);
	}

	/* Get current LP solution... */
	if (_MYCPX_solution (lp, &status, &z, x, NULL, NULL, NULL) NE 0) {
		fatal ("try_branch: Bug 2.");
	}

	/* Determine type of LP result... */
	switch (status) {
	case CPX_OPTIMAL:
	case CPX_OPTIMAL_INFEAS:
		break;

	case CPX_INFEASIBLE:
	case CPX_UNBOUNDED:
			/* (CPLEX 3.0 sometimes gives us infeasible!) */
	case CPX_OBJ_LIM:	/* Objective limit exceeded in Phase II. */
		z = ival;
		break;

	default:
		printf (" %% Status = %d\n", status);
		_MYCPX_lpwrite (lp, "core.lp");
		fatal ("try_branch: Bug 2.");
		break;
	}

	b_bd [0] = 0.0;
	b_bd [1] = 1.0;
	if (_MYCPX_chgbds (lp, 2, b_index, b_lu, b_bd) NE 0) {
		fatal ("try_branch: Bug 3.");
	}

	/* Restore the basis... */
	status = _MYCPX_loadbase (lp, basp -> cstat, basp -> rstat);
	if (status NE 0) {
		fprintf (stderr, "try_branch: status = %d\n", status);
		fatal ("try_branch: Bug 4.");
	}

	return (z);
}

#endif

/*
 * This routine computes the lower-bound for the current node, which
 * consists of solving the LP and generating violated constraints
 * until either:
 *
 *	- LP becomes infeasible
 *	- LP objective meets or exceeds cutoff value
 *	- LP solution is integral
 *	- separation finds no more violated constraints
 */

	static
	int
compute_good_lower_bound (

struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
int			i;
int			j;
int			n;
int			num_const;
int			status;
bitmap_t *		tmap;
bitmap_t *		fset_mask;
struct cinfo *		cip;
LP_t *			lp;
double *		x;
double			z;
struct constraint *	cp;
struct constraint *	tmp;
int			iteration;
int			fix_status;
int			num_fractional;
cpu_time_t		Tlp;
cpu_time_t *		Tp;
bool			is_int;
char			tbuf1 [16];
char			tbuf2 [80];
char			time_str [80];
char			title [80];
cpu_time_t		Tn [20];

	cip	  = bbip -> cip;
	tmap	  = bbip -> tmap;
	fset_mask = bbip -> fset_mask;
	lp	  = bbip -> lp;
	x	  = bbip -> x;

	Tp = &Tn [0];

	*Tp++ = get_cpu_time ();

	iteration = 1;
	num_const = 0;

	for (;;) {
		status = solve_LP_over_constraint_pool (bbip);
		z = bbip -> z;

		Tlp = get_cpu_time ();

		update_lp_solution_history (bbip);

		++(bbip -> node -> iter);
		++(bbip -> statp -> num_lps);

#if 0
		/* Display LP solution vector in machine-readable form... */
		for (i = 0; i < cip -> num_edges; i++) {
			printf (" %% %08lx %08lx\n",
				((bitmap_t *) &x [i]) [0],
				((bitmap_t *) &x [i]) [1]);
		}
#endif

		if (Check_Root_Constraints AND
		    (bbip -> node -> depth EQ 0)) {
			fwrite (x,
				1,
				cip -> num_edges * sizeof (*x),
				rcfile);
		}

#if 1
		convert_cpu_time (Tlp - *--Tp, time_str);
		while (Tp > &Tn [0]) {
			--Tp;
			convert_cpu_time (Tp [1] - Tp [0], tbuf1);
			(void) sprintf (tbuf2, "%s/%s", tbuf1, time_str);
			strcpy (time_str, tbuf2);
		}
		(void) sprintf (title,
				"Node %d LP %d Solution, length = %f, %s %d",
				bbip -> node -> num, bbip -> node -> iter,
				z, time_str, num_const);
#if 0
		plot_lp_solution (title, x, cip, BIG_PLOT);
#else
		(void) printf ("  %% %s\n", title);
#endif
#endif

		switch (status) {
		case BBLP_OPTIMAL:
			if (z >= bbip -> best_z) {
				bbip -> z = bbip -> best_z;
				return (LB_CUTOFF);
			}
			break;

		case BBLP_CUTOFF:
			bbip -> z = bbip -> best_z;
			return (LB_CUTOFF);

		case BBLP_INFEASIBLE:
			bbip -> z = bbip -> best_z;
			return (LB_INFEASIBLE);

		default:
			printf ("%% solve status = %d\n", status);
			fatal ("compute_good_lower_bound: Bug 3.");
		}

		/* Now get rid of any rows that have become	*/
		/* slack.  (We don't lose these constraints:	*/
		/* they're still sitting around in the		*/
		/* constraint pool.)				*/
		delete_slack_rows_from_LP (bbip);

		/* Solution is feasible, check for integer-feasible... */
		is_int = integer_feasible_solution (x,
						    tmap,
						    fset_mask,
						    cip,
						    &num_fractional);

		printf (" %% %d fractional variables\n", num_fractional);

		if (is_int) {
			/* All vars are either 0 or 1 and the	*/
			/* solution is connected -- we have a	*/
			/* Steiner tree!			*/

			/* Re-calculate the final objective	*/
			/* function to eliminate numerical	*/
			/* errors in the value of Z...		*/
			z = 0.0;
			for (i = 0; i < cip -> num_edges; i++) {
				if (x [i] + FUZZ < 1.0) continue;
				z += cip -> cost [i];
			}
			bbip -> z = z;
			if (z >= bbip -> best_z) {
				/* probably a repeat performance... */
				bbip -> z = bbip -> best_z;
				return (LB_CUTOFF);
			}
			bbip -> node -> optimal = TRUE;
			return (LB_INTEGRAL);
		}

#if 0
		/* NOTE -- this technique cannot correctly be	*/
		/* applied until an OPTIMAL solution to the LP	*/
		/* relaxation of the IP problem has been found!	*/
		/* Trying to fix variables while cuts are still	*/
		/* being generated can incorrectly fix vars and	*/
		/* cause the correct optimal solution to be	*/
		/* missed!					*/

		/* Try to fix some variables using reduced costs... */
		fix_status = reduced_cost_var_fixing (bbip);
		if (fix_status EQ VFIX_INFEASIBLE) {
			bbip -> z = bbip -> best_z;
			return (LB_INFEASIBLE);
		}
#endif

		/* Check to see if this node's objective value	*/
		/* is now high enough to be preempted...	*/
		if (bbip -> z > bbip -> preempt_z) {
			/* Node's value is no longer the lowest...	*/
			/* Preempt this one in favor of another.	*/
			return (LB_PREEMPTED);
		}

		/* Perhaps we have a new lower bound? */
		new_lower_bound (z, bbip);

		Tp = &Tn [0];
		*Tp++ = get_cpu_time ();

#if 1
		compute_heuristic_upper_bound (bbip);

		/* If we have improved the upper bound, it is possible	*/
		/* that this node can now be cutoff...			*/
		if (bbip -> z >= bbip -> best_z) {
			bbip -> z = bbip -> best_z;
			return (LB_CUTOFF);
		}
#endif

		if (force_branch_flag) {
			/* User kicked us!  Stop separating and branch! */
			force_branch_flag = FALSE;
			break;
		}


		if (bbip -> prevlb >= Trial_Branch_Threshold) {
			/* Test the effectiveness of each	*/
			/* fractional variable as a branch	*/
			/* variable.  Currently this produces	*/
			/* only debugging output, nothing else.	*/
			test_branching_variables (bbip);
		}

		/* Apply all separation algorithms to solution... */
		cp = do_separations (bbip, &Tp);

		if (cp EQ NULL) {
			/* No more violated constraints found! */
#if 0
			/* Try to fix some variables using reduced costs... */
			fix_status = reduced_cost_var_fixing (bbip);
			if (fix_status EQ VFIX_INFEASIBLE) {
				bbip -> z = bbip -> best_z;
				return (LB_INFEASIBLE);
			}
#endif
			break;
		}

		/* Add new contraints to the constraint pool. */
		num_const = add_constraints (bbip, cp);

		while (cp NE NULL) {
			tmp = cp;
			cp = tmp -> next;
			free ((char *) (tmp -> mask));
			free ((char *) tmp);
		}
		++iteration;
	}

#if 1
	/* Print execution times of final iteration... */
	convert_cpu_time (0, time_str);
	--Tp;
	while (Tp > &Tn [0]) {
		--Tp;
		convert_cpu_time (Tp [1] - Tp [0], tbuf1);
		(void) sprintf (tbuf2, "%s/%s", tbuf1, time_str);
		strcpy (time_str, tbuf2);
	}
	(void) printf ("  %% Final iteration: %s\n", time_str);
#endif

	/* Only get here with fractional solution and	*/
	/* no more violated constraints were found.	*/

	return (LB_FRACTIONAL);
}

/*
 * This routine copies the current LP solution into the current node's
 * recent history of LP solutions.
 */

	static
	void
update_lp_solution_history (

struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
int			nedges;
int			index;
struct bbnode *		nodep;
double *		srcx;
double *		dstx;

	nodep	= bbip -> node;
	nedges	= bbip -> cip -> num_edges;

	/* Get index of next history entry to fill... */
	index	= nodep -> xh_index;
	nodep -> xh_index = (index + 1) % NUM_X_HISTORIES_PER_NODE;

	dstx	= &(nodep -> xhist [index * nedges]);
	srcx	= bbip -> x;

	/* Copy LP solution into proper history entry. */
	memcpy (dstx, srcx, nedges * sizeof (*dstx));
}

/*
 * Routine to print out an updated lower bound value.
 */

	static
	void
new_lower_bound (

double			lb,		/* IN - new lower bound value */
struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
int		i;
struct cinfo *	cip;
double		prev;
double		old_gap;
double		new_gap;
cpu_time_t	t1;
char		buf1 [32];

	cip = bbip -> cip;

	prev = bbip -> prevlb;
	if ((prev > 0.0) AND (lb <= prev)) {
		/* There has been no improvement - get out. */
		return;
	}

	if (prev <= 0.0) {
		/* Don't make lower bound jump from zero... */
		prev = lb;
	}

	/* Print out the old and new lower bounds, with timestamp. */
	t1 = get_cpu_time ();
	convert_cpu_time (t1 - bbip -> t0, buf1);

	if ((bbip -> best_z >= DBL_MAX) OR (bbip -> best_z EQ 0.0)) {
		old_gap = 99.9;
		new_gap = 99.9;
	}
	else {
		old_gap = 100.0 * (bbip -> best_z - prev) / bbip -> best_z;
		new_gap = 100.0 * (bbip -> best_z - lb) / bbip -> best_z;
	}

	printf (" %% @LO %s %24.20f %2.10f\n",
		buf1, prev / cip -> scale_value, old_gap);
	printf (" %% @LN %s %24.20f %2.10f\n",
		 buf1, lb / cip -> scale_value, new_gap);

	bbip -> prevlb = lb;
}

/*
 * Routine to print out an updated upper bound value.
 */

	void
new_upper_bound (

double			ub,		/* IN - new upper bound value */
struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
struct cinfo *	cip;
double		prev;
int		i;
cpu_time_t	t1;
double		old_gap;
double		new_gap;
char		buf1 [64];
char		buf2 [64];

	cip = bbip -> cip;

	prev = bbip -> best_z;
	if (ub >= prev) {
		/* Supposed to be an improvement! */
		fatal ("new_upper_bound: Bug 1.");
	}

	/* We HAVE a new best solution! */
	bbip -> best_z = ub;

#if CPLEX
	{ double toobig, toosmall;
	  /* Set new cutoff value for future LPs... */
	  if (_MYCPX_setobjulim (ub, &toosmall, &toobig) NE 0) {
		fatal ("new_upper_bound: Bug 2.");
	  }
	}
#endif

#if LPSOLVE
	/* Set new cutoff value for future LPs... */
	/* (This may not really work in lp_solve.) */
	bbip -> lp -> obj_bound = ub;
#endif

	cut_off_existing_nodes (ub, bbip -> bbtree);

	/* Might want to do this if all other nodes were cut off. */
	update_node_preempt_value (bbip);

	/* Now print out the trace messages. */
	if (prev >= DBL_MAX) {
		/* Don't make upper bound jump from infinity... */
		prev = ub;
	}

	/* Print out the old and new lower and upper bounds, with timestamp. */
	t1 = get_cpu_time ();
	convert_cpu_time (t1 - bbip -> t0, buf1);

	if (bbip -> prevlb EQ 0.0) {
		old_gap = 99.9;
		new_gap = 99.9;
	}
	else {
		old_gap = 100.0 * (prev - bbip -> prevlb) / prev;
		new_gap = 100.0 * (ub - bbip -> prevlb) / ub;
	}

	printf (" %% @UO %s %24.20f %2.10f\n",
		buf1, prev / cip -> scale_value, old_gap);
	printf (" %% @UN %s %24.20f %2.10f\n",
		buf1, ub / cip -> scale_value, new_gap);
}

/*
 * This routine deletes any existing node whose objective value is
 * cut off by the given latest feasible integer solution.
 */

	static
	void
cut_off_existing_nodes (

double		best_z,		/* IN - new best objective value */
struct bbtree *	tp		/* IN - branch-and-bound tree */
)
{
int		num_cut;
struct bbheap *	hp;
struct bbnode *	p;

	num_cut = 0;

	/* We process the nodes from WORST to best... */
	hp = &(tp -> heap [WORST_NODE_HEAP]);

	while (hp -> nheap > 0) {
		/* Get node with highest objective value... */
		p = hp -> array [0];
		if (p -> index [WORST_NODE_HEAP] NE 0) {
			fatal ("cut_off_existing_nodes: Bug 1.");
		}
		if (p -> z < best_z) {
			/* All remaining nodes are < best_z... */
			break;
		}

		/* This node has been cut off! */
		delete_node_from_bbtree (p, tp);
		p -> next = tp -> free;
		tp -> free = p;
		++num_cut;
	}

	if (num_cut > 0) {
		printf (" %%	=== %d nodes cut off ===\n", num_cut);
	}
}

/*
 * This routine updates the node preemption value.  The node to be
 * preempted must be active when this routine is called (i.e., it must
 * be removed from the heap so that the "next best" node is at the top
 * of the heap).
 */

	static
	void
update_node_preempt_value (

struct bbinfo *		bbip		/* IN - branch-and-bound info */
)
{
struct bbtree *		tp;
struct bbheap *		hp;
struct bbnode *		node2;

	tp = bbip -> bbtree;
	hp = &(tp -> heap [BEST_NODE_HEAP]);

	if (hp -> nheap <= 0) {
		/* No other nodes.  Preempt only at cutoff. */
		bbip -> preempt_z = bbip -> best_z;
	}
	else {
		/* Preempt current node when next-best is exceeded. */
		node2 = hp -> array [0];
		bbip -> preempt_z = node2 -> z;
	}
}

/*
 * This routine performs most of the separations -- in the proper order.
 */

	static
	struct constraint *
do_separations (

struct bbinfo *		bbip,		/* IN - branch-and-bound info */
cpu_time_t **		Tpp		/* IN/OUT - CPU time vector */
)
{
double *		x;
bitmap_t *		vert_mask;
bitmap_t *		edge_mask;
struct cinfo *		cip;
struct comp *		comp;
struct comp *		p;
struct comp **		hookp;
struct comp *		p2;
cpu_time_t *		Tp;
struct constraint *	cp;
struct constraint *	cp2;
struct constraint *	tmp;
bool			print_flag;
bool			optimal;
bool			any_skipped;

	x		= bbip -> x;
	vert_mask	= bbip -> tmap;
	edge_mask	= bbip -> fset_mask;
	cip		= bbip -> cip;

	Tp = *Tpp;

	/* Find all zero-weight cutsets... */
	cp = find_cutset_constraints (x, vert_mask, edge_mask, cip);
	*Tp++ = get_cpu_time ();

	/* Find solid integer cycles... */
	cp = find_integer_cycles (x, vert_mask, edge_mask, cp, cip);
	*Tp++ = get_cpu_time ();

	/* Break problem up into congested components... */
	print_flag = TRUE;
	comp = find_congested_components (x,
					  vert_mask,
					  edge_mask,
					  print_flag,
					  cip);

	/* Exhaustively enumerate all components that are sufficiently	*/
	/* small...  Delete them from the list when done.		*/
	hookp = &comp;
	while ((p = *hookp) NE NULL) {
		if (p -> num_terms <= SEC_ENUM_LIMIT) {
			cp = enumerate_all_subtours (p, cp, bbip -> node);
			*hookp = p -> next;
			p -> next = NULL;
			free_congested_component (p);
		}
		else {
			hookp = &(p -> next);
		}
	}
	*Tp++ = get_cpu_time ();

	/* Find violated SEC's using a heuristic flow	*/
	/* formulation.					*/
	for (p = comp; p NE NULL; p = p -> next) {
		p -> cp = sec_flow_heuristic (p,
					      x,
					      edge_mask,
					      cip,
					      p -> cp);
	}
	*Tp++ = get_cpu_time ();

	/* Find small-cardinality subtour violations	*/
	/* by partial enumeration...			*/
	for (p = comp; p NE NULL; p = p -> next) {
		p -> cp = find_small_subtours (p,
					       p -> cp,
					       bbip -> node);
	}
	*Tp++ = get_cpu_time ();

	/* Discard each component for which we have found at least one	*/
	/* violation.  Gather all constraints onto the main list...	*/
	hookp = &comp;
	for (;;) {
		p = *hookp;
		if (p EQ NULL) break;
		cp2 = p -> cp;
		if (cp2 NE NULL) {
			/* Gather these constraints onto main list... */
			p -> cp = NULL;
			while (cp2 NE NULL) {
				tmp = cp2 -> next;
				cp2 -> next = cp;
				cp = cp2;
				cp2 = tmp;
			}
			*hookp = p -> next;
			p -> next = NULL;
			free_congested_component (p);

		}
		else {
			/* No constraints yet for this component.  We	*/
			/* may want to try the more expensive method...	*/
			/* Retain this component.			*/
			hookp = &(p -> next);
		}
	}

	/* Time to use the new-fangled SEC separator... */
	p2 = comp;
	comp = NULL;
	cp = sec_flow_separator (&p2, x, edge_mask, bbip, cp);
	*Tp++ = get_cpu_time ();

	/* If this separation routine does not find any SEC violations,	*/
	/* it means that none exist!					*/
	optimal = TRUE;

	/* Note: congested components are all freed now... */

#if 0
	if (cp EQ NULL) {
		/* Nothing else found -- look for fractional cutsets... */
		cp = find_fractional_cutsets (x,
					      bbip -> csip,
					      vert_mask,
					      edge_mask,
					      cip);
		*Tp++ = get_cpu_time ();
	}
#endif

	if ((cp EQ NULL) AND optimal) {
		/* We KNOW that we have NO violations!  The LP	*/
		/* relaxation is now OPTIMAL!			*/
		bbip -> node -> optimal = TRUE;
	}

	*Tpp = Tp;

	return (cp);
}

/*
 * This routine attempts to use LP reduced costs to fix variables.  Any
 * variable whose reduced cost exceeds the current LP/IP gap can be
 * permanently fixed to its current value (either 0 or 1) for the duration
 * of the current bb-node (and all of its children).
 */

	static
	int
reduced_cost_var_fixing (

struct bbinfo *		bbip	/* IN - branch-and-bound info */
)
{
int			i;
int			nedges;
int			nmasks;
int			status;
int			nfix0;
int			nfix1;
struct cinfo *		cip;
LP_t *			lp;
double *		rc;
double			gap;
double			c;
int *			newfix0;
int *			newfix1;

	if (bbip -> best_z >= DBL_MAX) {
		/* Can't reasonably attempt this until we have	*/
		/* a valid upper bound...			*/
		return (VFIX_NOTHING_FIXED);
	}
	gap = bbip -> best_z - bbip -> z;

	/* Only fix if we significantly exceed the gap... */
	gap *= (1.0 + FUZZ);

	lp	= bbip -> lp;
	cip	= bbip -> cip;
	nedges	= cip -> num_edges;
	nmasks	= cip -> num_edge_masks;

	newfix0 = NEWA (nedges, int);
	newfix1 = NEWA (nedges, int);

	nfix0	= 0;
	nfix1	= 0;

	rc = bbip -> dj;

	for (i = 0; i < nedges; i++) {
		c = rc [i];
		if (c < 0) {
			c = -c;
		}
		if (c > gap) {
			if (bbip -> x [i] < 0.5) {
				newfix0 [nfix0++] = i;
			}
			else {
				newfix1 [nfix1++] = i;
			}
		}
	}

	status = fix_variables (bbip, newfix0, nfix0, newfix1, nfix1);

	free ((char *) newfix1);
	free ((char *) newfix0);

	return (status);
}

/*
 * This routine fixes variables to zero and/or one.  We are given two
 * lists of variables to fixed, those to be fixed to zero, and those
 * to be fixed to one.  This routine then iteratively applies a series
 * of deductive steps that can cause additional variables to be fixed
 * based upon connectivity and compatibility criteria.
 */

	static
	int
fix_variables (

struct bbinfo *		bbip,		/* IN - branch-and-bound info */
int *			fix_to_0,	/* IN - vars to fix to 0 */
int			nfix0,		/* IN - number of vars fixed to 0 */
int *			fix_to_1,	/* IN - vars to fix to 1 */
int			nfix1		/* IN - number of vars fixed to 1 */
)
{
int			i;
int			j;
int			k;
int			t;
int			fs;
int			nedges;
int			nmasks;
int			kmasks;
int			status;
int			last_fset;
struct cinfo *		cip;
bitmap_t *		fixmask0;
bitmap_t *		fixmask1;
bitmap_t *		terms_checked;
int *			ep1;
int *			ep2;
int *			ep3;
int *			ep4;
int *			vp1;
int *			vp2;
int			vars_fixed;

#define	PRINT_FIXED_VARIABLES

	if ((nfix0 <= 0) AND (nfix1 <= 1)) {
		/* Nothing to do, get out quickly. */
		return (VFIX_NOTHING_FIXED);
	}

	cip	= bbip -> cip;
	nedges	= cip -> num_edges;
	nmasks	= cip -> num_edge_masks;
	kmasks	= cip -> num_vert_masks;

	fixmask0 	= NEWA (2 * nmasks + kmasks, bitmap_t);
	fixmask1 	= fixmask0 + nmasks;
	terms_checked	= fixmask1 + nmasks;

	/* Initialize masks of vars in fix-to-0 and fix-to-1 lists. */
	/* We use these to prevent adding duplicate entries later on. */
	for (i = 0; i < nmasks; i++) {
		fixmask0 [i] = 0;
		fixmask1 [i] = 0;
	}
	for (i = 0; i < nfix0; i++) {
		SETBIT (fixmask0, fix_to_0 [i]);
	}
	for (i = 0; i < nfix1; i++) {
		SETBIT (fixmask1, fix_to_1 [i]);
	}

	status = VFIX_NOTHING_FIXED;

	/* Iteratively fix variables until no more fixing can be done. */
	do {
		vars_fixed = FALSE;

		/* =============== Handle fixing vars to 0 =============== */

		ep1 = fix_to_0;
		ep2 = ep1;
		ep3 = ep1 + nfix0;
		while (ep1 < ep3) {
			fs = *ep1++;
			CLRBIT (fixmask0, fs);
			if (NOT BITON (bbip -> fset_mask, fs)) continue;
			if (BITON (bbip -> fixed, fs)) {
				if (NOT BITON (bbip -> value, fs)) {
					/* Already fixed to zero!  Ignore. */
					continue;
				}
				/* Already fixed to one! */
				status = VFIX_INFEASIBLE;
				goto alldone;
			}
			/* Fix it to zero now! */
			SETBIT (bbip -> fixed, fs);
			CLRBIT (bbip -> value, fs);
			change_var_bounds (bbip -> lp, fs, 0.0, 0.0);
#ifdef PRINT_FIXED_VARIABLES
			printf (" %%	Fixed x%-3d = 0\n", fs);
#endif
			/* Save this edge for later check. */
			*ep2++ = fs;
			/* We have fixed at least 1 variable! */
			status = VFIX_VARIABLES_FIXED;
			vars_fixed = TRUE;
		}
		ep1 = fix_to_0;
		if (ep1 < ep2) {
			/* Check if any of the vars we just set to 0	*/
			/* permit us to deduce vars that must be 1...	*/
			for (i = 0; i < kmasks; i++) {
				terms_checked [i] = 0;
			}
			while (ep1 < ep2) {
				i = *ep1++;
				vp1 = cip -> edge [i];
				vp2 = cip -> edge [i + 1];
				while (vp1 < vp2) {
					t = *vp1++;
					if (BITON (terms_checked, t)) continue;
					SETBIT (terms_checked, t);
					ep3 = cip -> term_trees [t];
					ep4 = cip -> term_trees [t + 1];
					k = 0;
					last_fset = -1;
					while (ep3 < ep4) {
						fs = *ep3++;
						if (NOT BITON (bbip -> fset_mask, fs)) continue;
						if (BITON (bbip -> fixed, fs) AND
						    NOT BITON (bbip -> value, fs)) continue;
						/* Full set fs has term t */
						/* and is NOT fixed to zero. */
						last_fset = fs;
						++k;
						if (k > 1) break;
					}
					if (k <= 0) {
						/* disconnected terminal! */
						status = VFIX_INFEASIBLE;
						goto alldone;
					}
					if ((k EQ 1) AND
					    NOT BITON (fixmask1, last_fset)) {
						/* one full set left, it */
						/* must be taken! */
						SETBIT (fixmask1, last_fset);
						fix_to_1 [nfix1++] = last_fset;
					}
				}
			}
		}
		/* Set the Fix-to-0 list to empty.  Fixmask0 should now	*/
		/* have all bits turned off.				*/
		nfix0 = 0;


		/* =============== Handle fixing vars to 1 =============== */

		ep1 = fix_to_1;
		ep2 = ep1 + nfix1;
		while (ep1 < ep2) {
			fs = *ep1++;
			CLRBIT (fixmask1, fs);
			if (NOT BITON (bbip -> fset_mask, fs)) continue;
			if (BITON (bbip -> fixed, fs)) {
				if (BITON (bbip -> value, fs)) {
					/* Already fixed to one!  Ignore. */
					continue;
				}
				/* Already fixed to zero! */
				status = VFIX_INFEASIBLE;
				goto alldone;
			}
			/* Fix it to one now! */
			SETBIT (bbip -> fixed, fs);
			SETBIT (bbip -> value, fs);
			change_var_bounds (bbip -> lp, fs, 1.0, 1.0);
#ifdef PRINT_FIXED_VARIABLES
			printf (" %%	Fixed x%-3d = 1\n", fs);
#endif
			/* We have fixed at least 1 variable! */
			status = VFIX_VARIABLES_FIXED;
			vars_fixed = TRUE;

			/* Fix every *other* incompatible FST to zero! */
			ep3 = cip -> inc_edges [fs];
			ep4 = cip -> inc_edges [fs + 1];
			while (ep3 < ep4) {
				j = *ep3++;
				if (j EQ fs) continue;
				if (NOT BITON (fixmask0, j)) {
					SETBIT (fixmask0, j);
					fix_to_0 [nfix0++] = j;
				}
			}
		}
		/* Set the Fix-to-1 list to empty.  Fixmask1 should now	*/
		/* have all bits turned off.				*/
		nfix1 = 0;
	} while (vars_fixed);

alldone:

	free ((char *) fixmask0);

	return (status);
}

/*
 * This routine changes the bounds on the given LP variable...
 */

	static
	void
change_var_bounds (

LP_t *			lp,		/* IN - LP to changes bounds of */
int			var,		/* IN - variable to fix */
double			lower,		/* IN - lower bound */
double			upper		/* IN - upper bound */
)
{
#if CPLEX
int			b_index [2];
char			b_lu [2];
double			b_bd [2];

	b_index [0] = var;	b_lu [0] = 'L';		b_bd [0] = lower;
	b_index [1] = var;	b_lu [1] = 'U';		b_bd [1] = upper;
	if (_MYCPX_chgbds (lp, 2, b_index, b_lu, b_bd) NE 0) {
		fatal ("change_var_bounds: Bug 1.");
	}
#endif

#if LPSOLVE
	set_bounds (lp, var + 1, lower, upper);
#endif
}

/*
 * This routine checks to see if we have an integer feasible solution.
 * First we check for integrality, then we check connectedness.
 */

	static
	bool
integer_feasible_solution (

double *		x,		/* IN - LP solution to check. */
bitmap_t *		tmap,		/* IN - subset of terminals. */
bitmap_t *		fset_mask,	/* IN - subset of full-sets. */
struct cinfo *		cip,		/* IN - compatibility info. */
int *			num_fractional	/* OUT - number of fractional vars. */
)
{
int			i;
int			t;
int			fs;
int			fs2;
int			kmasks;
int			nedges;
int			nmasks;
int			num_int;
int			num_frac;
int			starting_fset;
struct pset *		terms;
bitmap_t *		integral_fsets;
int *			ep1;
int *			ep2;
int *			vp1;
int *			vp2;
int *			sp;
int *			stack;
bitmap_t		mask;
bitmap_t *		terms_left;

	nedges = cip -> num_edges;
	nmasks = cip -> num_edge_masks;
	kmasks = cip -> num_vert_masks;

	/* First do a quick check of the integrality of the solution... */
	num_frac	= 0;
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		if (x [i] <= FUZZ) continue;
		if (x [i] + FUZZ >= 1.0) continue;

		/* Variable has a fractional value... */
		++num_frac;
	}
	*num_fractional = num_frac;

	if (num_frac > 0) {
		/* We have fractional variables -- solution is	*/
		/* definitely NOT integer feasible...		*/
		return (FALSE);
	}

	/* All solution variables are either 0 or 1 -- integral!  This	*/
	/* case is much less common.  Loop back over the solution,	*/
	/* making a note of all full sets present in the solution.	*/

	integral_fsets = NEWA (nmasks, bitmap_t);
	for (i = 0; i < nmasks; i++) {
		integral_fsets [i] = 0;
	}

	num_int		= 0;
	starting_fset	= -1;
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		if (x [i] + FUZZ >= 1.0) {
			SETBIT (integral_fsets, i);
			starting_fset = i;
			++num_int;
		}
	}

	if (starting_fset < 0) {
		/* No full sets in solution -- problem must have one or	*/
		/* fewer terminals!  This is connected by default.	*/
		free ((char *) integral_fsets);
		return (TRUE);
	}

	/* Create temporary mask of terminals we have not yet seen... */
	terms_left = NEWA (kmasks, bitmap_t);
	for (i = 0; i < kmasks; i++) {
		terms_left [i] = tmap [i];
	}

	stack = NEWA (num_int, int);
	sp = stack;

	/* Find connected component containing the starting_fset... */
	CLRBIT (integral_fsets, starting_fset);
	--num_int;
	*sp++ = starting_fset;

	while (sp > stack) {
		fs = *--sp;
		vp1 = cip -> edge [fs];
		vp2 = cip -> edge [fs + 1];
		while (vp1 < vp2) {
			t = *vp1++;
			if (NOT BITON (terms_left, t)) continue;
			CLRBIT (terms_left, t);
			ep1 = cip -> term_trees [t];
			ep2 = cip -> term_trees [t + 1];
			while (ep1 < ep2) {
				fs2 = *ep1++;
				if (NOT BITON (integral_fsets, fs2)) continue;
				CLRBIT (integral_fsets, fs2);
				--num_int;
				*sp++ = fs2;
			}
		}
	}

	/* See if any terminals were not reached... */
	mask = 0;
	for (i = 0; i < kmasks; i++) {
		mask |= terms_left [i];
	}

	free ((char *) stack);
	free ((char *) terms_left);
	free ((char *) integral_fsets);

	if (mask NE 0) {
		/* At least one more connected component -- solution	*/
		/* is not connected, and therefore infeasible!  (We	*/
		/* also know we have at least one integer cycle!)	*/
		return (FALSE);
	}

	/* Solution is a Steiner tree!  (Not necessarily minimal.) */

	return (TRUE);
}

/*
 * This routine goes through each of the constraints in the LP tableaux
 * for the root node.  Each constraint is checked against each LP
 * solution (recorded in the rcfile) to find the earliest iteration in
 * which we could have had root LP optimality -- if our separation
 * algorithm magically generated the *right* constraints.
 */

	static
	void
check_root_constraints (

struct bbinfo *		bbip	/* IN - branch-and-bound info */
)
{
int			i;
int			j;
int			nedges;
int			nbytes;
int			iter;
int *			list;
struct cinfo *		cip;
struct cpool *		pool;
int *			ip1;
int *			ip2;
int *			ip3;
double *		x;

	cip	= bbip -> cip;
	pool	= bbip -> cpool;

	/* Develop the list of all (non-initial) constraints... */
	list = NEWA (pool -> nlprows, int);
	ip3 = list;
	for (i = 0; i < pool -> nlprows; i++) {
		j = pool -> lprows [i];
		if (j < pool -> initrows) continue;
		*ip3++ = j;
	}


	fclose (rcfile);

	rcfile = fopen ("/tmp/lp.x", "r");

	nedges = cip -> num_edges;
	nbytes = nedges * sizeof (double);

	x = NEWA (nedges, double);

	iter = 0;
	while (ip3 > list) {
		i = fread (x, 1, nbytes, rcfile);
		if (i NE nbytes) {
			fatal ("check_root_constraints: Bug 1.");
		}

		/* Delete all remaining constraints that violate x. */
		ip1 = ip2 = list;
		while (ip2 < ip3) {
			j = *ip2++;
			if (NOT is_violation (pool -> rows [j].coefs, x)) {
				/* No violation -- keep constraint around. */
				*ip1++ = j;
			}
		}
		ip3 = ip1;
		printf (" %% @r iter %d, %d constraints left\n",
			iter, ip3 - list);
		if (ip3 <= list) break;
		++iter;
	}

	printf (" %% @RC Could have gotten root constraints in %d iterations!\n", iter);

	fclose (rcfile);
	rcfile = NULL;

	free ((char *) x);
	free ((char *) list);
}
