/***********************************************************************

	File:	sec2.c
	Rev:	a-1
	Date:	05/16/97

	Copyright (c) 1997, 1999 by David M. Warme

************************************************************************

	Deterministic separation procedure for the "generalized SEC's"
	(Subtour Elimination Constraints).  This method reduces the
	problem to min-cut on a simple bipartite network.

************************************************************************

	Modification Log:

	a-1:	05/16/97	warme
		: Created.

************************************************************************/

#include "bb.h"
#include "constrnt.h"
#include "flow.h"
#include "genps.h"
#include "sec2.h"
#include "sec_comp.h"
#include "sec_heur.h"
#include "steiner.h"


/*
 * Global Routines
 */

struct constraint *	sec_flow_separator (struct comp **,
					    double *,
					    bitmap_t *,
					    struct bbinfo *,
					    struct constraint *);


/*
 * External References
 */

	/* None */


/*
 * Local Types
 */

struct sec_flow_info {

	/* Data used by the flow solver... */
	struct flow_prob	prob;	/* The network flow formulation */
	struct flow_soln	soln;	/* The network flow solution */
	struct flow_temp	temp;	/* Temporary data structures */
};



/*
 * Local Routines
 */

static void			build_SEC_flow_formulation (
					struct comp *		comp,
					int			t,
					struct sec_flow_info *	flowp);
static double			do_flow_problem (struct comp *	comp,
						 int		t,
						 bitmap_t *	S);
static void			free_SEC_flow_formulation (
					struct sec_flow_info *	flowp);


/*
 * Local Variables
 */

	/* none */

/*
 * This is the main routine for the NEW deterministic separation
 * procedure for the generalized Subtour Elimination Constraints.  This
 * method reduces the problem to min-cut on a simple bipartite network.
 *
 * We take a chain of congested components.  We separate and then destroy
 * each congested component in the chain.
 */

	struct constraint *
sec_flow_separator (

struct comp **		comp_hookp,	/* IN/OUT - congested component(s) */
double *		x,		/* IN - the LP solution to separate */
bitmap_t *		fset_mask,	/* IN - subset of valid full sets */
struct bbinfo *		bbip,		/* IN - branch-and-bound info */
struct constraint *	cp		/* IN - existing constraints */
)
{
int			i;
int			t;
int			kmasks;
struct comp *		comp;
struct cinfo *		cip;
bitmap_t *		bp1;
double			z;
bitmap_t *		S;

	cip = bbip -> cip;

#if 0
	plot_lp_solution ("LP solution to separate", x, cip, BIG_PLOT);
#endif

	S = NEWA (cip -> num_vert_masks, bitmap_t);

	for (;;) {
		comp = *comp_hookp;
		if (comp EQ NULL) break;
		if (comp -> num_terms <= 0) {
			/* Free up this component... */
			*comp_hookp = comp -> next;
			comp -> next = NULL;
			free_congested_component (comp);
			continue;
		}
		if (comp -> num_terms EQ 1) {
			/* Because of the total-degree constraint, this	 */
			/* cannot happen unless something is very wrong! */
			fatal ("sec_flow_separator: Bug 1.");
		}

		if (comp -> num_terms <= SEC_ENUM_LIMIT) {
			/* We have whittled this component down to	*/
			/* something small.  Use brute force on the	*/
			/* rest...					*/
			cp = enumerate_all_subtours (comp, cp, bbip -> node);

			/* Free up this component... */
			*comp_hookp = comp -> next;
			comp -> next = NULL;
			free_congested_component (comp);
			continue;
		}

		/* Find the LEAST congested terminal.  this is the one	*/
		/* we are going to try to force into the solution,	*/
		/* since that is the one we would like to delete from	*/
		/* the set afterward...					*/
		t = find_least_congested_terminal (NULL, comp);

#if 0
		kmasks = comp -> kmasks;
		bp1 = &(comp -> tmasks [t * kmasks]);
		for (i = 0; i < cip -> num_terms; i++) {
			if (BITON (bp1, i)) break;
		}
		printf (" %% -------------------------"
			    "-------------------------\n"
			" %% separating comp with %d terms, %d trees,"
			    " forcing terminal %d\n"
			" %% -------------------------"
			    "-------------------------\n",
			comp -> num_terms, comp -> num_fsets, i);
#endif

		/* Find worst SEC violation involving terminal t. */
		z = do_flow_problem (comp, t, S);

#if 0
		print_mask (" %% S =", S, comp -> num_terms);
		printf (" %%	f(S) = %-24.15g\n", z);
#endif

		if (z < (1.0 - FUZZ)) {
			/* Add new violated constraint to the list... */
			cp = check_component_subtour (S,
						      comp,
						      cp,
						      x,
						      fset_mask,
						      cip);
		}

		/* We have found the worst violation (if any) involving	*/
		/* terminal t.  We can now eliminate t from further	*/
		/* consideration...					*/

		*comp_hookp = delete_terminal_from_component (t, comp);
	}

	free ((char *) S);

	return (cp);
}

/*
 * This routine performs a single flow sub-problem.  We are given a
 * terminal to FORCE into the solution.  We find the worst SEC violation
 * involving that terminal.
 */

	static
	double
do_flow_problem (

struct comp *		comp,		/* IN - congested component */
int			t,		/* IN - terminal to force */
bitmap_t *		S		/* OUT - a most-violated subtour */
)
{
int			i;
int			j;
int			k;
int			size;
int *			ip1;
int *			ip2;
double			sum;
double			z;
struct sec_flow_info	flow_info;

	build_SEC_flow_formulation (comp, t, &flow_info);

	compute_max_flow (&flow_info.prob,
			  &flow_info.temp,
			  &flow_info.soln);

	/* Construct the solution.  These are the terminals that are	*/
	/* on the FAR side of the cut.  Also, compute the z = f(S)	*/
	/* value to return to our caller.				*/
	k = BMAP_ELTS (comp -> num_terms);
	for (i = 0; i < k; i++) {
		S [i] = 0;
	}
	size = 0;
	for (i = 0; i < comp -> num_terms; i++) {
		if (BITON (flow_info.soln.cut, i)) continue;
		SETBIT (S, i);
		++size;
	}

	free_SEC_flow_formulation (&flow_info);

	sum = 0.0;
	for (i = 0; i < comp -> num_fsets; i++) {
		ip1 = comp -> fsterms [i];
		ip2 = comp -> fsterms [i + 1];
		j = 0;
		while (ip1 < ip2) {
			k = *ip1++;
			if (BITON (S, k)) {
				++j;
			}
		}
		if (j >= 2) {
			sum += (j - 1) * (comp -> x [i]);
		}
	}

	z = ((double) size) - sum;

	return (z);
}

/*
 * This routine builds the network flow formulation for the
 * deterministic separation procedure.
 */

	static
	void
build_SEC_flow_formulation (

struct comp *		comp,		/* IN - congested component */
int			t,		/* IN - terminal to force */
struct sec_flow_info *	flowp		/* OUT - SEC flow formulation */
)
{
int			i;
int			j;
int			k;
int			nterms;
int			ntrees;
int			nmasks;
int			num_arcs;
int			num_nodes;
int			fs;
int			arc_num;
int			num_used_fsets;
int			first_fset_node;
int *			used_fsets;
bitmap_t *		unused_fsmask;
struct flow_prob *	prob;
int *			ip1;
int *			ip2;
int *			outlist;
int *			inlist;
int *			srcp;
int *			dstp;
double *		capp;
int *			counts;
int **			ptrs;
double			sum;

	nterms	= comp -> num_terms;
	ntrees	= comp -> num_fsets;

	/* Compute a list of all component full sets that	*/
	/* DO NOT contain the terminal "t".			*/
	nmasks		= BMAP_ELTS (ntrees);
	unused_fsmask	= NEWA (nmasks, bitmap_t);
	for (i = 0; i < nmasks; i++) {
		unused_fsmask [i] = 0;
	}
	num_used_fsets = ntrees;
	ip1 = comp -> tfsets [t];
	ip2 = comp -> tfsets [t + 1];
	while (ip1 < ip2) {
		fs = *ip1++;
		SETBIT (unused_fsmask, fs);
		--num_used_fsets;
	}

	used_fsets = NEWA (num_used_fsets, int);
	ip1 = used_fsets;
	for (i = 0; i < ntrees; i++) {
		if (BITON (unused_fsmask, i)) continue;
		*ip1++ = i;
	}
	free ((char *) unused_fsmask);

	if (ip1 NE (used_fsets + num_used_fsets)) {
		/* lost count somewhere? */
		fatal ("build_SEC_flow_formulation: Bug 1.");
	}

	/* Tally up the total number of nodes and arcs needed	*/
	/* for the flow graph.  For the sake of	simplicity, we	*/
	/* include a node for t, but there will	be no arcs	*/
	/* associated with it...				*/

	/* One node per terminal.  One source and one sink	*/
	/* node.  One node per USED full set.			*/
	num_nodes	= nterms + 2 + num_used_fsets;

	/* One arc per terminal, one arc per USED full set,	*/
	/* plus one arc for each terminal of every USED full	*/
	/* set.							*/
	num_arcs	= nterms + num_used_fsets;
	for (i = 0; i < num_used_fsets; i++) {
		fs = used_fsets [i];
		num_arcs += (comp -> fsterms [fs + 1] - comp -> fsterms [fs]);
	}

	/* Start filling in the flow problem instance... */
	prob = &(flowp -> prob);
	prob -> num_nodes	= num_nodes;
	prob -> num_arcs	= num_arcs;

	/* Assign node numbers for the source and sink nodes... */
	prob -> source	= nterms;
	prob -> sink	= nterms + 1;
	first_fset_node	= nterms + 2;

	/* Now that we know how big the directed flow graph is,	*/
	/* allocate storage for the various data structures...	*/

	prob -> out		= NEWA (num_nodes + 1, int *);
	prob -> in		= NEWA (num_nodes + 1, int *);
	prob -> arc_src		= NEWA (num_arcs, int);
	prob -> arc_dst		= NEWA (num_arcs, int);
	prob -> capacity	= NEWA (num_arcs, double);

	outlist			= NEWA (num_arcs, int);
	inlist			= NEWA (num_arcs, int);

	/* Generate the arcs from the source node to each USED full set. */

	srcp = prob -> arc_src;
	dstp = prob -> arc_dst;
	capp = prob -> capacity;
	arc_num	= 0;
	for (i = 0; i < num_used_fsets; i++) {
		fs = used_fsets [i];
		j = first_fset_node + i;
		*srcp++		= prob -> source;
		*dstp++		= j;
		*capp++		= comp -> x [fs];
		++arc_num;

		/* Generate the arcs from each full set node to the	*/
		/* corresponding terminal nodes.  These all have weight	*/
		/* 2, which is essentially infinite.			*/
		ip1 = comp -> fsterms [fs];
		ip2 = comp -> fsterms [fs + 1];
		while (ip1 < ip2) {
			k = *ip1++;
			*srcp++		= j;
			*dstp++		= k;
			*capp++		= 2.0;
			++arc_num;
		}
	}

	free ((char *) used_fsets);

	/* Now generate one arc from each terminal node to the sink	*/
	/* node.  These all have weight (Bi - 1), where Bi is the	*/
	/* congestion level of terminal i.				*/
	for (i = 0; i < nterms; i++) {
		sum = -1.0;
		ip1 = comp -> tfsets [i];
		ip2 = comp -> tfsets [i + 1];
		while (ip1 < ip2) {
			fs = *ip1++;
			sum += comp -> x [fs];
		}
		*srcp++		= i;
		*dstp++		= prob -> sink;
		*capp++		= sum;
		++arc_num;
	}

	if (arc_num NE num_arcs) {
		fatal ("build_SEC_flow_formulation: Bug 2.");
	}

	/* We have now specified the directed flow graph as a	*/
	/* list of directed arcs.  Time to construct the	*/
	/* adjacency lists -- for each node we build a list of	*/
	/* outgoing and incoming arc numbers.  Do the outgoing	*/
	/* lists first...					*/

	counts	= NEWA (num_nodes, int);
	ptrs	= NEWA (num_nodes, int *);

	for (i = 0; i < num_nodes; i++) {
		counts [i] = 0;
	}
	for (i = 0; i < num_arcs; i++) {
		++(counts [prob -> arc_src [i]]);
	}
	ip1 = outlist;
	for (i = 0; i < num_nodes; i++) {
		ptrs [i] = ip1;
		prob -> out [i] = ip1;
		ip1 += counts [i];
	}
	prob -> out [i] = ip1;
	for (i = 0; i < num_arcs; i++) {
		j = prob -> arc_src [i];
		ip1 = ptrs [j]++;
		*ip1 = i;
	}

	/* Now do the incoming arc lists... */
	for (i = 0; i < num_nodes; i++) {
		counts [i] = 0;
	}
	for (i = 0; i < num_arcs; i++) {
		++(counts [prob -> arc_dst [i]]);
	}
	ip1 = inlist;
	for (i = 0; i < num_nodes; i++) {
		ptrs [i] = ip1;
		prob -> in [i] = ip1;
		ip1 += counts [i];
	}
	prob -> in [i] = ip1;
	for (i = 0; i < num_arcs; i++) {
		k = prob -> arc_dst [i];
		ip1 = ptrs [k]++;
		*ip1 = i;
	}

	/* Free temporary memory used to build things... */
	free ((char *) counts);
	free ((char *) ptrs);

	/* Initialize the buffers used to hold flow solutions */
	/* and temporary data... */
	create_flow_solution_data (prob, &(flowp -> soln));
	create_flow_temp_data (prob, &(flowp -> temp));
}

/*
 * This routine frees up the memory allocated by the given SEC max-flow
 * formulation.
 */

	static
	void
free_SEC_flow_formulation (

struct sec_flow_info *	flowp		/* IN - SEC flow formulation */
)
{
	/* Free up the buffers used to hold flow solutions */
	/* and temporary data... */
	free_flow_temp_data (&(flowp -> temp));
	free_flow_solution_data (&(flowp -> soln));

	/* Free up the problem formulation... */
	free ((char *) (flowp -> prob.in [0]));
	free ((char *) (flowp -> prob.out [0]));

	free ((char *) (flowp -> prob.capacity));
	free ((char *) (flowp -> prob.arc_dst));
	free ((char *) (flowp -> prob.arc_src));
	free ((char *) (flowp -> prob.in));
	free ((char *) (flowp -> prob.out));
}
