/***********************************************************************

	File:	sll.c
	Rev:	a-1
	Date:	11/10/98

	Copyright (c) 1998, 1999 by Martin Zachariasen

************************************************************************

	Implementation of a variant of the Smith, Lee and Liebman
	heuristic for Euclidean Steiner trees (Networks 11, 1981)

************************************************************************

	Modification Log:

	a-1:	11/10/98	martinz
		: Created.  Derived from C++ geosteiner96 program

************************************************************************/

#include "dsuf.h"
#include "steiner.h"

#define ANSI_DECLARATORS
#define REAL coord_t
#include "triangle.h"


/*
 * Global Routines
 */

dist_t		smith_lee_liebman (struct pset *);


/*
 * Local Constants
 */

#define PI (3.141592653589793238462643383279502884197169399375105820974944592308)

#define HALF_SQRT3	(0.86602540378443864676)


/*
 * Local Types
 */

struct sll_edge {
	dist_t		len;
	pnum_t		p1;
	pnum_t		p2;
	bool		mst;
};

struct sll_fst {
	dist_t		len;
	dist_t		ratio;
	pnum_t		p [4];
};


/*
 * Local Routines
 */

static int		comp_edges (const void *, const void *);


/*
 * Local Variables
 */

static struct sll_edge *	sll_edges;
static struct sll_fst  *	sll_fsts;

/*
 * Does the sequence p1, p2, p3 make a left turn?
 */

	static
	bool
right_turn (

struct point *		p1,	/* IN - first point */
struct point *		p2,	/* IN - second point */
struct point *		p3	/* IN - third point */
)
{
	return ( (p1 -> x - p2 -> x) * (p1 -> y - p3 -> y)
		 <
		 (p1 -> y - p2 -> y) * (p1 -> x - p3 -> x) );
}


/*
 * Does the sequence p1, p2, p3 make a right turn?
 */

	static
	bool
left_turn (

struct point *		p1,	/* IN - first point */
struct point *		p2,	/* IN - second point */
struct point *		p3	/* IN - third point */
)
{
	return ( (p1 -> x - p2 -> x) * (p1 -> y - p3 -> y)
		 >
		 (p1 -> y - p2 -> y) * (p1 -> x - p3 -> x) );
}

/*
 * Square of distance between points p1 and p2
 */

	static
	dist_t
sqr_dist (

struct point *		p1,	/* IN - first point */
struct point *		p2	/* IN - second point */
)
{
double		dx, dy;

	dx = p1 -> x - p2 -> x;
	dy = p1 -> y - p2 -> y;
	return (dx*dx + dy*dy);
}


/*
 * Find equilateral point for points p1 and p2
 */

	static
	void
eq_point (

struct point *		p1,	/* IN - first point */
struct point *		p2,	/* IN - second point */
struct point *		e	/* OUT - equilateral point */
)
{
double		dx, dy;

	dx = p1 -> x - p2 -> x;
	dy = p1 -> y - p2 -> y;

	e -> x = p2 -> x + 0.5*dx - HALF_SQRT3*dy;
	e -> y = p2 -> y + 0.5*dy + HALF_SQRT3*dx;
}


/*
 * Find center of equilateral circle
 */

	static
	void
eq_circle_center (

struct point *		p1,	/* IN - first point on circle */
struct point *		p2,	/* IN - second second on circle */
struct point *		e,	/* IN - third point on circle */
struct point *		c	/* OUT - center of circle */
)
{
	c -> x = (p1 -> x + p2 -> x + e -> x) / 3.0;
	c -> y = (p1 -> y + p2 -> y + e -> y) / 3.0;
}

/*
 * Test if the points p1,p2,p3 make up a wedge
 * that is greater than 120 degrees
 */

	static
	bool
wedge120 (

struct point *		p1,	/* IN - first point */
struct point *		p2,	/* IN - second point */
struct point *		p3	/* IN - third point */
)
{
struct point	e;
struct point	c;

	if (left_turn (p1,p3,p2)) {
		eq_point (p1, p3, &e);
		eq_circle_center (p1, p3, &e, &c);
	}
	else {
		eq_point (p3, p1, &e);
		eq_circle_center (p3, p1, &e, &c);
	}

	if (sqr_dist (&c, p2) <= sqr_dist (&c, p1)) {
		/* p2 is inside circle -> angle greater than 120 deg */
		return (TRUE);
	}

	/* less than 120 deg */

	return (FALSE);
}

/*
 * Intersection between two line segments pq and rs.
 * Adapted from Dave's more generic intersection code.
 */

	static
	bool
segment_intersection (

struct point *		p,	/* IN - first endpoint of first segment */
struct point *		q,	/* IN - second endpoint of first segment */
struct point *		r,	/* IN - first endpoint of second segment */
struct point *		s,	/* IN - second endpoint of second segment */
struct point *		is	/* OUT - intersection point */
)
{
	/* This routine takes two line segments PQ and RS and determines
	   their intersection (if any).  It makes use of a coordinate
	   transformation to reduce the number of special cases to a
	   minimum.  Line segment PQ runs from point P to point Q, and
	   line segment RS runs from point R to point S.  We create a
	   parametric form for line segment PQ as:

		XPQ(alpha) = (1 - alpha) * P.x + alpha * Q.x
		YPQ(alpha) = (1 - alpha) * P.y + alpha * Q.y

	   As alpha varies from 0.0 to 1.0, the parametric point follows
	   line segment PQ from P to Q.  Similarly, we parameterize line
	   segment RS as a function of beta:

		XRS(beta) = (1 - beta) * R.x + beta * S.x
		YRS(beta) = (1 - beta) * R.y + beta * S.y

	   The (infinite) lines containing segments PQ and RS intersect when:

		[XPQ(alpha) = XRS(beta)] AND [YPQ(alpha) = YRS(beta)]

	   Note that if (0 <= alpha <= 1) AND (0 <= beta <= 1) then the
	   line segments actually intersect, not just the lines.  This is
	   the really nifty property of this coordinate transformation!

	   Expanding out the simultaneous equations gives:

		(1 - alpha) * P.x + alpha * Q.x = (1 - beta) * R.x + beta * S.x
		(1 - alpha) * P.y + alpha * Q.y = (1 - beta) * R.y + beta * S.y

	   or:

		(Q.x - P.x) * alpha + (R.x - S.x) * beta = (R.x - P.x)
		(Q.y - P.y) * alpha + (R.y - S.y) * beta = (R.y - P.y)

	   which we write as:

		A * alpha + B * beta = C
		D * alpha + E * beta = F

	   Unless this system is singular, its solutions are:

			 B*F - C*E		C*D - A*F
		alpha = -----------	beta = -----------
			 B*D - A*E		B*D - A*E

	   The system is singular if either line segment is zero-length, or
	   if the lines have the same slope (parallel or colinear).  These
	   cases are handled separately.
	*/

double		alpha, beta;
double		A, B, C, D, E, F;
double		denom;
struct point *	rp;
struct point *	sp;
struct point *	tmp;

	A = q -> x - p -> x;
	B = r -> x - s -> x;
	C = r -> x - p -> x;
	D = q -> y - p -> y;
	E = r -> y - s -> y;
	F = r -> y - p -> y;

	rp = r;
	sp = s;

	denom = B * D - A * E;

	if (denom < 0) {
		/* switch points R and S... */
		tmp = rp;
		rp = sp;
		sp = tmp;

		denom = - denom;
		B = - B;
		C = rp -> x - p -> x;
		E = - E;
		F = rp -> y - p -> y;
	}

	if (denom NE 0.0) {
		/* Easy case -- lines are NOT parallel. */
		alpha = (B * F - C * E);
		if ((alpha < 0.0) OR (denom < alpha)) {
			/* First segment does not reach to second */
			/* line segment. */
			return (FALSE);
		}

		beta = C * D - A * F;
		if ((beta < 0.0) OR (denom < beta)) {
			/* Second line segment does not reach to first */
			/* line segment. */
			return (FALSE);
		}

		/* We have a single point of intersection! */

		alpha /= denom;
		is -> x = (1.0 - alpha) * p -> x + alpha * q -> x;
		is -> y = (1.0 - alpha) * p -> y + alpha * q -> y;
		return (TRUE);
	}

	/* The two line segments are parallel.
	   Report no (unique) intersection point. */

	return (FALSE);
}

/*
 * Compute the length of a shortest full Steiner tree
 * for a set of terminals with up to 4 terminals.
 *  Special fast version which only computes length.
 * Returns 0.0 if no Steiner tree exists.
 * Assume terminals are ordered as they appear on Steiner polygon.
 */

	static
	double
fst_length (

struct pset *		pts,	/* IN - terminal list */
int *			terms	/* IN - indices of terminals to consider */
)
{
int			i;
int			term_count;
struct point *		a;
struct point *		b;
struct point *		c;
struct point *		d;
struct point		e, ctr, e_ad, e_cb;
double			l;
double			min_length;

	/* What is the number of terminals? (should be four or less) */
	term_count = 0;
	while ((term_count < 4) AND (terms [term_count] >= 0)) {
		++term_count;
	}

	if (term_count EQ 2) {
		return (EDIST (&(pts -> a [terms [0]]),
			       &(pts -> a [terms [1]])));
	}

	if (term_count EQ 3) {
		a = &(pts -> a [terms [0]]);
		b = &(pts -> a [terms [1]]);
		c = &(pts -> a [terms [2]]);

		eq_point (a, b, &e);
		eq_circle_center (a, b, &e, &ctr);

		if (right_turn (&e, a, c) AND
		    left_turn (&e, b, c) AND
		    (sqr_dist (&ctr, c) > sqr_dist (&ctr, a))) {
			return EDIST (&e, c);
		}
	}

	if (term_count EQ 4) {
		min_length = INF_DISTANCE;
		for (i = 0; i <= 1; i++) {
			/* Using Lemma 5.2 p. 64 in Hwang, Richards, Winter */
			a = &(pts -> a [terms [i]]);
			d = &(pts -> a [terms [i+1]]);
			c = &(pts -> a [terms [i+2]]);
			b = &(pts -> a [terms [(i+3) % 4]]);

			/* Find intersetion point between ac and bd.
			   It is the same in both iterations */

			if ((i EQ 0) AND
			    NOT segment_intersection (a, c, b, d, &ctr)) break;

			eq_point (a, d, &e_ad);
			eq_point (c, b, &e_cb);

			if (NOT wedge120 (d, a, &e_cb) AND
			    NOT wedge120 (&e_cb, d, a) AND
			    NOT wedge120 (&e_ad, b, c) AND
			    NOT wedge120 (b, c, &e_ad) AND
			    NOT wedge120 (a, &ctr, d)) {
				l = EDIST (&e_ad, &e_cb);
				if (l < min_length) {
					min_length = l;
				}
			}
		}
		if (min_length < INF_DISTANCE) return (min_length);
	}

	return (0.0);
}

/*
 * For sorting edges by length (should be changed!!)
 */

	static
	int
comp_edges (

const void *		p1,
const void *		p2
)
{
double			l1;
double			l2;

	l1 = sll_edges [ *(int*) p1].len;
	l2 = sll_edges [ *(int*) p2].len;

	if (l1 < l2) return (-1);
	if (l1 > l2) return (1);
	return (0);
}


/*
 * For sorting FSTs by ratio (should be changed!!)
 */

	static
	int
comp_fsts (

const void *		p1,
const void *		p2
)
{
double			l1;
double			l2;

	l1 = sll_fsts [ *(int*) p1].ratio;
	l2 = sll_fsts [ *(int*) p2].ratio;

	if (l1 < l2) return (-1);
	if (l1 > l2) return (1);
	return (0);
}

/*
 * Add generated FST to list of FSTs
 */

	static
	void
add_fst (

struct pset *	pts,		/* IN - terminal list */
int *		terms,		/* IN - indices of terminals in FST */
struct sll_fst * sll_fsts,	/* IN/OUT - list of FSTs */
int *		fst_count,	/* IN/OUT - current FST count */
double		mst_l		/* Length of MST spanning terminals in FST */
)
{
int		j;
double		smt_l;
double		ratio;

	smt_l = fst_length (pts, terms);

	if (smt_l > 0) {
		ratio = smt_l / mst_l;
		if (ratio < 1.0) {
			for (j = 0; j < 4; j++) {
				sll_fsts [*fst_count].p [j] = terms [j];
			}
			sll_fsts [*fst_count].len   = smt_l;
			sll_fsts [*fst_count].ratio = ratio;
			++(*fst_count);
		}
	}
}

/*
 * Variant of heuristic by Smith, Lee and Liebman (Networks 11, 1981)
 * All 4-terminal candidates with three connected MST-edges
 * are put on the priority queue
 */

	double
smith_lee_liebman (

struct pset *	pts	/* IN - terminals list for which
				heuristic tree should be constructed */
)
{
int			i, j, jj, ei, fi, fst_count, mst_count;
int			ni, nj, nt, np1, np2, p1, p2, p4;
bool			neighbour_edge_mst;
bool			nb_edge;
bool			convex_region;
bool			in_same_block;
int			neighbour_edge;
int			neighbour_edge_idx;
int			outgoing_mst_edge;
int			term_count;
int			root [4];
int			terms [4];
int *			triangleedges;
int *			sortededges;
int *			sortedfsts;
dist_t			mst_l, total_mst_l, outgoing_mst_edge_l;
dist_t			total_smt_l;
struct dsuf		mstsets, fstsets;
struct triangulateio	in, out, vorout;

	/* Special cases */

	if (pts -> n <= 1) return (0.0);
	if (pts -> n EQ 2) return (EDIST (&(pts -> a [0]),
					  &(pts -> a [1])));

	/* Transfer to triangulate data structure and call triangle */

	in.numberofpoints = pts -> n;
	in.pointlist = NEWA (2*in.numberofpoints, REAL);

	for (i = 0; i < pts -> n; i++) {
		in.pointlist [2*i  ] = pts -> a [i].x;
		in.pointlist [2*i+1] = pts -> a [i].y;
	}

	in.numberofpointattributes	= 0;
	in.pointattributelist		= NULL;
	in.pointmarkerlist		= NULL;
	in.numberofsegments		= 0;
	in.numberofholes		= 0;
	in.numberofregions		= 0;
	in.regionlist			= NULL;
	out.pointlist			= NULL;
	out.trianglelist		= NULL;
	out.neighborlist		= NULL;

	triangulate ("znNBQ", &in, &out, &vorout);

	if (out.numberoftriangles EQ 0) {
		/* If there are no triangles (all input points co-linear), */
		/* then quit */
		free (in.pointlist);
		free (out.pointlist);
		free (out.trianglelist);
		free (out.neighborlist);
		return (INF_DISTANCE);
	}

	/* Construct edge information */
	sll_edges	= NEWA (out.numberofedges, struct sll_edge);
	triangleedges	= NEWA (3*out.numberoftriangles, int);
	for (i = 0; i < 3*out.numberoftriangles; i++) {
		triangleedges [i] = -1;
	}

	ei = 0;
	for (i = 0; i < out.numberoftriangles; i++) {
		for (j = 0; j < 3; j++) {
			p1 = out.trianglelist [3*i + j];
			p2 = out.trianglelist [3*i + ((j+1) % 3)];
			if (triangleedges [3*i + j] EQ -1) {
				/* only add once */
				sll_edges [ei].len = EDIST (&(pts -> a [p1]),
							    &(pts -> a [p2]));
				sll_edges [ei].p1  = p1;
				sll_edges [ei].p2  = p2;
				sll_edges [ei].mst = FALSE;
				triangleedges [3*i + j] = ei;

				/* Now go through neighbouring triangles */
				/* and add information about the new edge */
				for (ni = 0; ni < 3; ni++) {
					nt = out.neighborlist [3*i + ni];
					if (nt EQ -1) continue;
					for (nj = 0; nj < 3; nj++) {
						np1 = out.trianglelist [3*nt + nj];
						np2 = out.trianglelist [3*nt + ((nj+1) % 3)];
						if ((np1 EQ p2) AND
						    (np2 EQ p1)) {
							 /* found reverse edge */
							triangleedges [3*nt + nj] = ei;
						}
					}
				}
				++ei;
			}
		}
	}

	/* Sort edges */

	sortededges = NEWA (out.numberofedges, int);
	for (i = 0; i < out.numberofedges; i++) {
		sortededges [i] = i;
	}
	qsort (sortededges, out.numberofedges, sizeof(int), comp_edges);

	/* Use Kruskal to find MST */

	dsuf_create (&mstsets, pts -> n);
	for (i = 0; i < pts -> n; i++) {
		dsuf_makeset (&mstsets, i);
	}
	total_mst_l = 0.0;
	for (i = 0; i < out.numberofedges; i++) {
		/* go through edges in sorted order */
		ei = sortededges [i];
		root [1] = dsuf_find (&mstsets, sll_edges [ei].p1);
		root [2] = dsuf_find (&mstsets, sll_edges [ei].p2);

		if (root [1] NE root [2]) {
			dsuf_unite (&mstsets, root [1], root [2]);
			total_mst_l += sll_edges [ei].len;
			sll_edges [ei].mst = TRUE; /* remember this edge */
		}
	}

	/* Generate FSTs with 3 and 4 terminals */

	sll_fsts = NEWA (4*out.numberoftriangles, struct sll_fst);
	fst_count = 0;
	for (i = 0; i < out.numberoftriangles; i++) {
		/* Count the number of MST edges and find their length sum */
		mst_count = 0;
		mst_l = 0.0;
		for (j = 0; j < 3; j++) {
			ei = triangleedges [3*i + j];
			if (sll_edges [ei].mst) {
				++mst_count;
				mst_l += sll_edges [ei].len;
			}
		}

		/* If there are two MST edges try to construct an FST */
		if (mst_count NE 2) continue;

		for (j = 0; j < 3; j++) {
			terms [j] = out.trianglelist [3*i + j];
		}
		terms [3] = -1;

		/* Add 3-terminal FST */
		add_fst (pts, terms, sll_fsts, &fst_count, mst_l);

		/* Now go through neighbouring triangles and add */
		/* valid FSTs */
		for (ni = 0; ni < 3; ni++) {
			 /* get triangle index */
			nt = out.neighborlist [3*i + ni];
			if (nt EQ -1) continue;

			/* Find neighbouring edge and outgoing MST */
			/* edge (note that there can be at most one) */
			neighbour_edge		= -1;
			neighbour_edge_idx	= -1;
			outgoing_mst_edge	= -1;
			neighbour_edge_mst	= FALSE;
			outgoing_mst_edge_l	= 0.0;

			for (nj = 0; nj < 3; nj++) {
				ei = triangleedges [3*nt + nj];

				/* Is this the neighouring edge? */
				nb_edge = FALSE;
				for (j = 0; j < 3; j++) {
					if (triangleedges [3*i + j] EQ ei) {
						nb_edge = TRUE;
					}
				}
				if (nb_edge) {
					if (sll_edges [ei].mst) {
						/* neighbour edge is an */
						/* MST edge */
						neighbour_edge_mst = TRUE;
					}
					neighbour_edge	   = ei;
					neighbour_edge_idx = nj;
				}
				else if (sll_edges [ei].mst) {
					/* outgoing MST edge identified */
					outgoing_mst_edge = ei;
					outgoing_mst_edge_l
						= sll_edges [ei].len;
				}
			}

			if (outgoing_mst_edge < 0) continue;

			if (neighbour_edge_mst AND (i >= nt)) continue;
			/* only consider once... */

			/* Idenfify fourth terminal */

			p4 = out.trianglelist [3*nt + ((neighbour_edge_idx + 2) % 3)];

			/* Make list of points on border of triangles */

			j = -1;
			jj = -1;
			do {
				terms [++jj] = out.trianglelist [3*i + (++j)];
			} while (triangleedges [3*i + j] NE neighbour_edge);
			terms [++jj] = p4;
			while (j < 2) {
				terms [++jj] = out.trianglelist [3*i + (++j)];
			}

			/* Make sure that the two triangles make up a */
			/* convex region */

			convex_region = TRUE;
			for (j = 0; j < 4; j++) {
				if (right_turn (&(pts -> a [terms [j]]),
						&(pts -> a [terms [(j+1) % 4]]),
						&(pts -> a [terms [(j+2) % 4]]))) {
					convex_region = FALSE;
					break;
				}
			}

			if (convex_region) {
				/* Add 4-terminal FST */
				add_fst (pts,
					 terms,
					 sll_fsts,
					 &fst_count,
					 mst_l + outgoing_mst_edge_l);
			}
		}
	}

	/* Sort FSTs */

	sortedfsts = NEWA (fst_count, int);
	for (i = 0; i < fst_count; i++) {
		sortedfsts [i] = i;
	}
	qsort (sortedfsts, fst_count, sizeof (int), comp_fsts);

	/* Build heuristic tree */

	dsuf_create (&fstsets, pts -> n);
	for (i = 0; i < pts -> n; i++) {
		dsuf_makeset (&fstsets, i);
	}
	total_smt_l = 0.0;
	for (i = 0; i < fst_count; i++) {
		/* go through FSTs in sorted order */
		fi = sortedfsts [i];
		term_count = 0;
		while ((term_count < 4) AND
		       (sll_fsts [fi].p [term_count] >= 0)) {
			++term_count;
		}

		/* check that no two terminals are in the same block */
		in_same_block = FALSE;
		for (j = 0; j < term_count; j++) {
			root [j] = dsuf_find (&fstsets, sll_fsts [fi].p [j]);
		}
		for (j = 0; j < term_count-1; j++) {
			for (jj = j+1; jj < term_count; jj++) {
				if (root [j] EQ root [jj]) {
					in_same_block = TRUE;
					break;
				}
			}
		}
		if (in_same_block) continue;

		for (j = 1; j < term_count; j++) {
			dsuf_unite (&fstsets,
				    dsuf_find (&fstsets, sll_fsts [fi].p [0]),
				    dsuf_find (&fstsets, sll_fsts [fi].p [j]));
		}
		total_smt_l += sll_fsts [fi].len;
	}

	for (i = 0; i < out.numberofedges; i++) {
		/* go through edges in sorted order */
		ei = sortededges [i];
		root [1] = dsuf_find (&fstsets, sll_edges [ei].p1);
		root [2] = dsuf_find (&fstsets, sll_edges [ei].p2);
		if (root [1] NE root [2]) {
			dsuf_unite (&fstsets, root [1], root [2]);
			total_smt_l += sll_edges [ei].len;
		}
	}

	/* Free all allocated arrays, including those allocated by Triangle */

	dsuf_destroy (&mstsets);
	dsuf_destroy (&fstsets);

	free (in.pointlist);
	free (out.pointlist);
	free (out.trianglelist);
	free (out.neighborlist);
	free (triangleedges);
	free (sll_edges);
	free (sll_fsts);
	free (sortededges);
	free (sortedfsts);

	return (total_smt_l);
}
