/***********************************************************************

	File:	ub.c
	Rev:	a-2
	Date:	09/06/97

	Copyright (c) 1996, 1999 by David M. Warme

************************************************************************

	Heuristic for turning a fractional LP solution (a lower bound)
	into a feasible Steiner tree (an upper bound).  If such a
	Steiner tree can be found, and is the best seen so far, it
	becomes the new best feasible integer solution.

************************************************************************

	Modification Log:

	a-1:	07/17/96	warme
		: Created.
	a-2:	09/06/97	warme
		: Completely replaced this algorithm with a new one
		:  suggested by Martin Zachariasen.

************************************************************************/

#include "bb.h"
#include "dsuf.h"
#include "steiner.h"
#include "ub.h"


/*
 * Global Routines
 */

void			compute_heuristic_upper_bound (struct bbinfo *);
int *			compute_fstmst_index (struct cinfo *);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static void		sort_fst_list_by_lp_and_ratio (int *,
						       int,
						       double *,
						       int *);
static void		sort_fst_list_by_ratio (int *, int, int *);

/*
 * This routine pre-computes some useful information for the upper
 * bound heuristic -- namely the ranking of all FSTs in INCREASING
 * order by ratio of FST-length/MST-length.
 */

	int *
compute_fstmst_index (

struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			n;
int			h;
int			t1;
int			t2;
int *			p1;
int *			p2;
int *			p3;
int *			p4;
int *			endp;
int *			array;
int *			fstmst_index;
struct pset *		terms;
dist_t *		fst_len;
dist_t *		mst_len;
dist_t			key1;
dist_t			key2;
dist_t			key3;
dist_t			key4;

	n = cip -> num_edges;

	array = NEWA (n, int);

	/* Get FST lengths into an array... */
	fst_len = NEWA (n, dist_t);
	for (i = 0; i < n; i++) {
		fst_len [i] = cip -> cost [i];
	}

	/* Compute MST length of each FST... */
	mst_len = NEWA (n, dist_t);

	if ((cip -> metric EQ EUCLIDEAN) AND
	    (cip -> full_trees NE NULL)) {
		for (i = 0; i < n; i++) {
			terms = cip -> full_trees [i] -> terminals;
			/* Cheat and use rectilinear MST for now... */
			mst_len [i] = rect_mst_length (terms);
		}
	}
	else if ((cip -> metric EQ RECTILINEAR) AND
		 (cip -> full_trees NE NULL)) {
		for (i = 0; i < n; i++) {
			terms = cip -> full_trees [i] -> terminals;
			mst_len [i] = rect_mst_length (terms);
		}
	}
	else {
		/* Pretend each edge in the edge's MST has length 1! */
		for (i = 0; i < n; i++) {
			mst_len [i] = cip -> edge_size [i] - 1;
		}
	}

	for (i = 0; i < n; i++) {
		array [i] = i;
	}


	endp = &array [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &array [h];
		p1 = p4;
		while (p1 < endp) {
			t1 = *p1;
			key1 = fst_len [t1];
			key2 = mst_len [t1];
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				t2 = *p3;
				key3 = fst_len [t2];
				key4 = mst_len [t2];
				if (key3 * key2 <= key1 * key4) break;
				*p2 = t2;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = t1;
			++p1;
		}
	} while (h > 1);

	free ((char *) mst_len);
	free ((char *) fst_len);

	/* Produce array that ranks each FST... */
	fstmst_index = NEWA (n, int);

	for (i = 0; i < n; i++) {
		fstmst_index [array [i]] = i;
	}

	free ((char *) array);

	return (fstmst_index);
}

/*
 * This routine takes an existing LP solution (a fractional lower
 * bound for an optimal Steiner tree) and attempts to perturb it into
 * a feasible integer solution.  If this succeeds and we discover a
 * shorter solution than previously known, it becomes the current
 * best feasible integer solution.  Any existing nodes that exceed
 * this new upper bound are then cut off.
 *
 * We use a greedy heuristic: sort the valid hyperedges into decreasing
 * order by LP solution weight.  If two edges have the same LP solution
 * weight, the one with the smaller FST-length/MST-length ratio will be
 * placed first.
 *
 * Greedily add edges from this sorted list to an empty tree, discarding
 * any edge that would form a cycle.  This is just a variant of Kruskal's
 * algorithm for MST, extended for hypergraphs.
 *
 * The one real efficiency hack is that we use linear time to partition
 * the edges into those with weight 1, fractional weight, and weight 0.
 * We then sort each region independently.
 */

	void
compute_heuristic_upper_bound (

struct bbinfo *		bbip	/* IN - branch-and-bound info */
)
{
int			i;
int			j;
int			n;
int			e;
int			nverts;
int			nedges;
int *			ep1;
int *			ep2;
int *			ep3;
int *			vp1;
int *			vp2;
int *			vp3;
int *			vp4;
struct cinfo *		cip;
double *		x;
int *			edge_list;
int *			roots;
bool *			mark;
int			components;
dist_t			length;
struct dsuf		sets;

	cip		= bbip -> cip;
	x		= bbip -> x;

	nverts		= cip -> num_verts;
	nedges		= cip -> num_edges;

	/* Generate the list of valid full sets... */
	edge_list = NEWA (nedges, int);

	/* Put edges of weight 0.0 last, all the rest at the front. */
	ep1 = edge_list;
	ep2 = edge_list + nedges;
	for (i = 0; i < nedges; i++) {
		if (x [i] <= FUZZ) {
			*--ep2 = i;
		}
		else {
			*ep1++ = i;
		}
	}
	if (ep1 NE ep2) {
		fatal ("compute_heuristic_upper_bound: Bug 1.");
	}

	/* Sort the zero-weight ones only by FST/MST ratio. */
	ep1 = edge_list + nedges;
	sort_fst_list_by_ratio (ep2, ep1 - ep2, bbip -> fstmst_index);

	/* Now re-scan the non-zero edges, putting those of	*/
	/* weight 1.0 first.					*/
	ep1 = edge_list;
	ep3 = ep2;
	while (ep1 < ep2) {
		/* Scan forward from beginning for a fractional edge. */
		i = *ep1;
		if (x [i] + FUZZ >= 1.0) {
			++ep1;
			continue;
		}
		/* i is a fractional edge.  Scan backward from end	*/
		/* for an edge of weight 1.0.				*/
		for (;;) {
			--ep2;
			if (ep1 >= ep2) break;
			j = *ep2;
			if (x [j] + FUZZ >= 1.0) {
				/* i is fractional, j is integral -- swap. */
				*ep1 = j;
				*ep2 = i;
				++ep1;
				break;
			}
		}
	}
	if (ep1 NE ep2) {
		fatal ("compute_heuristic_upper_bound: Bug 2.");
	}

	/* Sort the integral FSTs by FST/MST ratio only. */
	ep1 = edge_list;
	sort_fst_list_by_ratio (ep1, ep2 - ep1, bbip -> fstmst_index);

	/* Sort the fractional FSTs by LP weight, then by FST/MST ratio. */
	sort_fst_list_by_lp_and_ratio (ep2,
				       ep3 - ep2,
				       x,
				       bbip -> fstmst_index);

	/* Initialize one disjoint subtree for each terminal. */
	dsuf_create (&sets, nverts);
	for (i = 0; i < nverts; i++) {
		dsuf_makeset (&sets, i);
	}

	mark = NEWA (nverts, bool);
	for (i = 0; i < nverts; i++) {
		mark [i] = FALSE;
	}
	roots = NEWA (nverts, int);

	/* Start the greedy Kruskal procedure... */
	components = nverts;
	length = 0.0;
	ep1 = edge_list;
	ep2 = edge_list;
	ep3 = edge_list + nedges;
	while (components > 1) {
		if (ep2 >= ep3) {
			/* FSTs ran out before tree constructed! */
			fatal ("compute_heuristic_upper_bound: Bug 3.");
		}
		e = *ep2++;
		vp3 = roots;
		vp1 = cip -> edge [e];
		vp2 = cip -> edge [e + 1];
		for (;;) {
			if (vp1 >= vp2) {
				/* No cycle!  Include e in solution! */
				*ep1++ = e;
				length += cip -> cost [e];
				/* Unite all subtrees joined, and clear	*/
				/* out the mark bits...			*/
				vp4 = roots;
				i = *vp4++;
				mark [i] = FALSE;
				while (vp4 < vp3) {
					j = *vp4++;
					mark [j] = FALSE;
					dsuf_unite (&sets, i, j);
					i = dsuf_find (&sets, i);
					--components;
				}
				break;
			}
			j = dsuf_find (&sets, *vp1++);
			if (mark [j]) {
				/* This FST would form a cycle. */
				while (roots < vp3) {
					mark [*--vp3] = FALSE;
				}
				break;
			}
			/* This FST terminal is in distinct subtree... */
			mark [j] = TRUE;
			*vp3++ = j;
		}
	}

	if (length < bbip -> best_z) {
		/* We have a better feasible integer solution! */
		for (i = 0; i < cip -> num_edge_masks; i++) {
			bbip -> smt [i] = 0;
		}
		while (ep1 > edge_list) {
			e = *--ep1;
			SETBIT (bbip -> smt, e);
		}

		new_upper_bound (length, bbip);
	}

	free ((char *) roots);
	free ((char *) mark);
	dsuf_destroy (&sets);
	free ((char *) edge_list);
}

/*
 * This routine sorts the full sets of integral weight (both 0.0 and 1.0).
 * In this case, only their FST-length/MST-length ratio is used.
 */

	static
	void
sort_fst_list_by_ratio (

int *		array,		/* IN - list of full set numbers to sort */
int		n,		/* IN - number of full sets to sort */
int *		fstmst_index	/* IN - ranking of FSTs by FST/MST ratio */
)
{
int		h;
int		t1;
int		t2;
int		key1;
int *		p1;
int *		p2;
int *		p3;
int *		p4;
int *		endp;

	endp = &array [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &array [h];
		p1 = p4;
		while (p1 < endp) {
			t1 = *p1;
			key1 = fstmst_index [t1];
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				t2 = *p3;
				if (fstmst_index [t2] <= key1) break;
				*p2 = t2;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = t1;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the fractional full sets.  The primary key is LP
 * solution weight, ties are broken using FST/MST ratio.
 */

	static
	void
sort_fst_list_by_lp_and_ratio (

int *		array,		/* IN - list of full set numbers to sort */
int		n,		/* IN - number of full sets to sort */
double *	x,		/* IN - LP solution weights */
int *		fstmst_index	/* IN - ranking of FSTs by FST/MST ratio */
)
{
int		h;
int		t1;
int		t2;
double		key1;
int *		p1;
int *		p2;
int *		p3;
int *		p4;
int *		endp;

	endp = &array [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &array [h];
		p1 = p4;
		while (p1 < endp) {
			t1 = *p1;
			key1 = x [t1];
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				t2 = *p3;
				if (x [t2] > key1) break;
				if ((x [t2] EQ key1) AND
				    (fstmst_index [t2] <=
				     fstmst_index [t1])) break;
				*p2 = t2;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = t1;
			++p1;
		}
	} while (h > 1);
}
